#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Implementation of a dependency tree."""

from __future__ import annotations

from typing import Container, Iterable, Optional

import numpy as np

from satpy import DataID, DatasetDict
from satpy.dataset import ModifierTuple, create_filtered_query
from satpy.dataset.data_dict import TooManyResults, get_key
from satpy.node import EMPTY_LEAF_NAME, CompositorNode, MissingDependencies, Node, ReaderNode
from satpy.utils import get_logger

LOG = get_logger(__name__)


class Tree:
    """A tree implementation."""

    # simplify future logic by only having one "sentinel" empty node
    # making it a class attribute ensures it is the same across instances
    empty_node = Node(EMPTY_LEAF_NAME)

    def __init__(self):
        """Set up the tree."""
        self._root = Node(None)

        # keep a flat dictionary of nodes contained in the tree for better
        # __contains__
        self._all_nodes = _DataIDContainer()

    def leaves(self,
               limit_nodes_to: Optional[Iterable[DataID]] = None,
               unique: bool = True
               ) -> list[Node]:
        """Get the leaves of the tree starting at the root.

        Args:
            limit_nodes_to: Limit leaves to Nodes with the names (DataIDs)
                specified.
            unique: Only include individual leaf nodes once.

        Returns:
            list of leaf nodes

        """
        if limit_nodes_to is None:
            return self._root.leaves(unique=unique)

        res = list()
        for child_id in limit_nodes_to:
            for sub_child in self._all_nodes[child_id].leaves(unique=unique):
                if not unique or sub_child not in res:
                    res.append(sub_child)
        return res

    def trunk(self,
              limit_nodes_to: Optional[Iterable[DataID]] = None,
              unique: bool = True,
              limit_children_to: Optional[Container[DataID]] = None,
              ) -> list[Node]:
        """Get the trunk nodes of the tree starting at this root.

        Args:
            limit_nodes_to: Limit searching to trunk nodes with the names
                (DataIDs) specified and the children of these nodes.
            unique: Only include individual trunk nodes once
            limit_children_to: Limit searching to the children with the specified
                names. These child nodes will be included in the result,
                but not their children.

        Returns:
            list of trunk nodes

        """
        if limit_nodes_to is None:
            return self._root.trunk(unique=unique,
                                    limit_children_to=limit_children_to)

        res = list()
        for child_id in limit_nodes_to:
            child_node = self._all_nodes[child_id]
            for sub_child in child_node.trunk(unique=unique, limit_children_to=limit_children_to):
                if not unique or sub_child not in res:
                    res.append(sub_child)
        return res

    def add_child(self, parent, child):
        """Add a child to the tree."""
        Node.add_child(parent, child)
        # Sanity check: Node objects should be unique. They can be added
        #               multiple times if more than one Node depends on them
        #               but they should all map to the same Node object.
        if self.contains(child.name):
            if self._all_nodes[child.name] is not child:
                raise RuntimeError
        if child is self.empty_node:
            # No need to store "empty" nodes
            return
        self._all_nodes[child.name] = child

    def add_leaf(self, ds_id, parent=None):
        """Add a leaf to the tree."""
        if parent is None:
            parent = self._root
        try:
            node = self[ds_id]
        except KeyError:
            node = Node(ds_id)
        self.add_child(parent, node)
        return node

    def __contains__(self, item):
        """Check if a item is in the tree."""
        return item in self._all_nodes

    def __getitem__(self, item):
        """Get an item of the tree."""
        return self._all_nodes[item]

    def contains(self, item):
        """Check contains when we know the *exact* DataID or DataQuery."""
        return super(_DataIDContainer, self._all_nodes).__contains__(item)

    def getitem(self, item):
        """Get Node when we know the *exact* DataID or DataQuery."""
        return super(_DataIDContainer, self._all_nodes).__getitem__(item)

    def __str__(self):
        """Render the dependency tree as a string."""
        return self._root.display()


class DependencyTree(Tree):
    """Structure to discover and store `Dataset` dependencies.

    Used primarily by the `Scene` object to organize dependency finding.
    Dependencies are stored used a series of `Node` objects which this
    class is a subclass of.

    """

    def __init__(self, readers, compositors=None, modifiers=None, available_only=False):
        """Collect Dataset generating information.

        Collect the objects that generate and have information about Datasets
        including objects that may depend on certain Datasets being generated.
        This includes readers, compositors, and modifiers.

        Composites and modifiers are defined per-sensor. If multiple sensors
        are available, compositors and modifiers are searched for in
        sensor alphabetical order.

        Args:
            readers (dict): Reader name -> Reader Object
            compositors (dict): Sensor name -> Composite ID -> Composite Object.
                Empty dictionary by default.
            modifiers (dict): Sensor name -> Modifier name -> (Modifier Class, modifier options).
                Empty dictionary by default.
            available_only (bool): Whether only reader's available/loadable
                datasets should be used when searching for dependencies (True)
                or use all known/configured datasets regardless of whether the
                necessary files were provided to the reader (False).
                Note that when ``False`` loadable variations of a dataset will
                have priority over other known variations.
                Default is ``False``.

        """
        super().__init__()
        self.readers = readers
        self.compositors = {}
        self.modifiers = {}
        self._available_only = available_only
        self.update_compositors_and_modifiers(compositors or {}, modifiers or {})

    def update_compositors_and_modifiers(self, compositors: dict, modifiers: dict) -> None:
        """Add additional compositors and modifiers to the tree.

        Provided dictionaries and the first sub-level dictionaries are copied
        to avoid modifying the input.

        Args:
            compositors (dict):
                Sensor name -> composite ID -> Composite Object
            modifiers (dict):
                Sensor name -> Modifier name -> (Modifier Class, modifier options)

        """
        for sensor_name, sensor_comps in compositors.items():
            self.compositors.setdefault(sensor_name, DatasetDict()).update(sensor_comps)
        for sensor_name, sensor_mods in modifiers.items():
            self.modifiers.setdefault(sensor_name, {}).update(sensor_mods)

    def copy(self):
        """Copy this node tree.

        Note all references to readers are removed. This is meant to avoid
        tree copies accessing readers that would return incompatible (Area)
        data. Theoretically it should be possible for tree copies to request
        compositor or modifier information as long as they don't depend on
        any datasets not already existing in the dependency tree.
        """
        new_tree = DependencyTree({}, self.compositors, self.modifiers)
        for c in self._root.children:
            c = c.copy(node_cache=new_tree._all_nodes)
            new_tree.add_child(new_tree._root, c)
        return new_tree

    def update_node_name(self, node, new_name):
        """Update 'name' property of a node and any related metadata."""
        old_name = node.name
        if old_name not in self._all_nodes:
            raise RuntimeError
        del self._all_nodes[old_name]
        node.update_name(new_name)
        self._all_nodes[new_name] = node

    def populate_with_keys(self, dataset_keys: set, query=None):
        """Populate the dependency tree.

        Args:
            dataset_keys (set): Strings, DataIDs, DataQuerys to find dependencies for
            query (DataQuery): Additional filter parameters. See
                              `satpy.readers.get_key` for more details.

        Returns:
            (Node, set): Root node of the dependency tree and a set of unknown datasets

        """
        unknown_datasets = list()
        known_nodes = list()
        for key in dataset_keys.copy():
            try:
                dsq = create_filtered_query(key, query)
                node = self._create_subtree_for_key(dsq, query)
            except MissingDependencies as unknown:
                unknown_datasets.append(unknown.missing_dependencies)
            else:
                known_nodes.append(node)
                self.add_child(self._root, node)

        for key in dataset_keys.copy():
            dataset_keys.discard(key)
        for node in known_nodes:
            dataset_keys.add(node.name)
        if unknown_datasets:
            raise MissingDependencies(unknown_datasets, "Unknown datasets:")

    def _create_subtree_for_key(self, dataset_key, query=None):
        """Find the dependencies for *dataset_key*.

        Args:
            dataset_key (str, float, DataID, DataQuery): Dataset identifier to locate
                                                         and find any additional
                                                         dependencies for.
            query (DataQuery): Additional filter parameters. See
                               `satpy.readers.get_key` for more details.

        """
        # 0 check if the *exact* dataset is already loaded
        try:
            node = self._get_subtree_for_existing_key(dataset_key)
        except MissingDependencies:
            # exact dataset isn't loaded, let's load it below
            pass
        else:
            return node

        # 1 try to get *best* dataset from reader
        try:
            node = self._create_subtree_from_reader(dataset_key, query)
        except TooManyResults:
            LOG.warning("Too many possible datasets to load for {}".format(dataset_key))
            raise MissingDependencies({dataset_key})
        except MissingDependencies:
            pass
        else:
            return node

        # 2 try to find a composite by name (any version of it is good enough)
        try:
            node = self._get_subtree_for_existing_name(dataset_key)
        except MissingDependencies:
            pass
        else:
            return node

        # 3 try to find a composite that matches
        try:
            node = self._create_subtree_from_compositors(dataset_key, query)
        except MissingDependencies:
            raise
        else:
            return node

    def _get_subtree_for_existing_key(self, dsq):
        try:
            node = self.getitem(dsq)
            LOG.trace("Found exact dataset already loaded: {}".format(node.name))
            return node
        except KeyError:
            LOG.trace("Exact dataset {} isn't loaded, will try reader...".format(dsq))
            raise MissingDependencies({dsq})

    def _create_subtree_from_reader(self, dataset_key, query):
        try:
            node = self._find_reader_node(dataset_key, query)
        except MissingDependencies:
            LOG.trace("Could not find dataset in reader: {}".format(dataset_key))
            raise
        else:
            LOG.trace("Found reader provided dataset:\n\tRequested: {}\n\tFound: {}".format(dataset_key, node.name))
            return node

    def _find_reader_node(self, dataset_key, query):  # noqa: D417
        """Attempt to find a `DataID` in the available readers.

        Args:
            dataset_key (str, float, DataID, DataQuery):
                Dataset name, wavelength, `DataID` or `DataQuery`
                to use in searching for the dataset from the
                available readers.

        """
        matching_ids = self._find_matching_ids_in_readers(dataset_key)
        unique_id = self._get_unique_matching_id(matching_ids, dataset_key, query)

        for reader_name, ids in matching_ids.items():
            if unique_id in ids:
                return self._get_unique_reader_node_from_id(unique_id, reader_name)

        raise RuntimeError("Data ID disappeared.")

    def _find_matching_ids_in_readers(self, dataset_key):
        matching_ids = {}
        for reader_name, reader_instance in self.readers.items():
            matching_ids[reader_name] = []
            try:
                ds_ids = reader_instance.get_dataset_key(dataset_key, available_only=self._available_only,
                                                         num_results=0, best=False)
            except KeyError:
                LOG.trace("Can't find dataset %s in reader %s", str(dataset_key), reader_name)
                continue
            matching_ids[reader_name].extend(ds_ids)
        return matching_ids

    def _get_unique_matching_id(self, matching_ids, dataset_key, query):
        """Get unique matching id from `matching_ids`, for a given `dataset_key` and some optional `query`."""
        all_ids = sum(matching_ids.values(), [])
        if len(all_ids) == 0:
            raise MissingDependencies({dataset_key})
        elif len(all_ids) == 1:
            result = all_ids[0]
        else:
            sorted_ids, distances = dataset_key.sort_dataids_with_preference(all_ids, query)
            try:
                result = self._get_unique_id_from_sorted_ids(sorted_ids, distances)
            except TooManyResults:
                LOG.trace("Too many datasets matching key {} in readers {}".format(dataset_key, matching_ids.keys()))
                raise TooManyResults("Too many keys matching: {}".format(dataset_key))
            except MissingDependencies:
                raise MissingDependencies({dataset_key})
        return result

    @staticmethod
    def _get_unique_id_from_sorted_ids(sorted_ids, distances):
        if distances[0] != np.inf:
            if distances[0] != distances[1]:
                result = sorted_ids[0]
            else:
                raise TooManyResults
        else:
            raise MissingDependencies
        return result

    def _get_unique_reader_node_from_id(self, data_id, reader_name):
        try:
            # now that we know we have the exact DataID see if we have already created a Node for it
            return self.getitem(data_id)
        except KeyError:
            # we haven't created a node yet, create it now
            return ReaderNode(data_id, reader_name)

    def _get_subtree_for_existing_name(self, dsq):
        try:
            # assume that there is no such thing as a "better" composite
            # version so if we find any DataIDs already loaded then
            # we want to use them
            node = self[dsq]
            LOG.trace("Composite already loaded:\n\tRequested: {}\n\tFound: {}".format(dsq, node.name))
            return node
        except KeyError:
            # composite hasn't been loaded yet, let's load it below
            LOG.trace("Composite hasn't been loaded yet, will load: {}".format(dsq))
            raise MissingDependencies({dsq})

    def _create_subtree_from_compositors(self, dataset_key, query):
        try:
            node = self._find_compositor(dataset_key, query)
            LOG.trace("Found composite:\n\tRequested: {}\n\tFound: {}".format(dataset_key, node and node.name))
        except KeyError:
            LOG.trace("Composite not found: {}".format(dataset_key))
            raise MissingDependencies({dataset_key})
        return node

    def _find_compositor(self, dataset_key, query):
        """Find the compositor object for the given dataset_key."""
        # NOTE: This function can not find a modifier that performs
        # one or more modifications if it has modifiers see if we can find
        # the unmodified version first

        if dataset_key.is_modified():
            implicit_dependency_node = self._create_implicit_dependency_subtree(dataset_key, query)
            dataset_key = self._promote_query_to_modified_dataid(dataset_key, implicit_dependency_node.name)
            try:
                compositor = self.get_modifier(dataset_key)
            except KeyError:
                raise KeyError("Can't find anything called {}".format(str(dataset_key)))
            compositor.attrs["prerequisites"] = [implicit_dependency_node] + list(compositor.attrs["prerequisites"])
        else:
            try:
                compositor = self.get_compositor(dataset_key)
            except KeyError:
                raise KeyError("Can't find anything called {}".format(str(dataset_key)))

        root = CompositorNode(compositor)
        composite_id = root.name

        prerequisite_filter = composite_id.create_filter_query_without_required_fields(dataset_key)

        # Get the prerequisites
        LOG.trace("Looking for composite prerequisites for: {}".format(dataset_key))
        prereqs = [create_filtered_query(prereq, prerequisite_filter) if not isinstance(prereq, Node) else prereq
                   for prereq in compositor.attrs["prerequisites"]]
        prereqs = self._create_required_subtrees(root, prereqs, query=query)
        root.add_required_nodes(prereqs)

        # Get the optionals
        LOG.trace("Looking for optional prerequisites for: {}".format(dataset_key))
        optionals = [create_filtered_query(prereq, prerequisite_filter) if not isinstance(prereq, Node) else prereq
                     for prereq in compositor.attrs["optional_prerequisites"]]
        optionals = self._create_optional_subtrees(root, optionals, query=query)
        root.add_optional_nodes(optionals)

        return root

    def _create_implicit_dependency_subtree(self, dataset_key, query):
        new_prereq = dataset_key.create_less_modified_query()
        src_node = self._create_subtree_for_key(new_prereq, query)
        return src_node

    def _promote_query_to_modified_dataid(self, query, dep_key):
        """Promote a query to an id based on the dataset it will modify (dep).

        Typical use case is requesting a modified dataset (query). This
        modified dataset most likely depends on a less-modified
        dataset (dep_key). The less-modified dataset must come from a reader
        (at least for now) or will eventually depend on a reader dataset.
        The original request key may be limited like
        (wavelength=0.67, modifiers=('a', 'b')) while the reader-based key
        should have all of its properties specified. This method updates the
        original request key so it is fully specified and should reduce the
        chance of Node's not being unique.

        """
        orig_dict = query._asdict()
        dep_dict = dep_key._asdict()
        for key, dep_val in dep_dict.items():
            # don't change the modifiers, just cast them to the right class
            if isinstance(dep_val, ModifierTuple):
                orig_dict[key] = dep_val.__class__(orig_dict[key])
            else:
                orig_dict[key] = dep_val
        return dep_key.from_dict(orig_dict)

    def get_compositor(self, key):
        """Get a compositor."""
        for sensor_name in sorted(self.compositors):
            try:
                return self.compositors[sensor_name][key]
            except KeyError:
                continue

        raise KeyError("Could not find compositor '{}'".format(key))

    def get_modifier(self, comp_id):
        """Get a modifer."""
        # create a DataID for the compositor we are generating
        modifier = comp_id["modifiers"][-1]
        for sensor_name in sorted(self.modifiers):
            modifiers = self.modifiers[sensor_name]
            compositors = self.compositors[sensor_name]
            if modifier not in modifiers:
                continue

            mloader, moptions = modifiers[modifier]
            moptions = moptions.copy()
            moptions.update(comp_id.to_dict())
            moptions["sensor"] = sensor_name
            compositors[comp_id] = mloader(_satpy_id=comp_id, **moptions)
            return compositors[comp_id]

        raise KeyError("Could not find modifier '{}'".format(modifier))

    def _create_required_subtrees(self, parent, prereqs, query=None):  # noqa: D417
        """Determine required prerequisite Nodes for a composite.

        Args:
            parent (Node): Compositor node to add these prerequisites under
            prereqs (Sequence): Strings (names), floats (wavelengths),
                                DataQuerys or Nodes to analyze.

        """
        prereq_nodes, unknown_datasets = self._create_prerequisite_subtrees(parent, prereqs, query)
        if unknown_datasets:
            raise MissingDependencies(unknown_datasets)
        return prereq_nodes

    def _create_optional_subtrees(self, parent, prereqs, query=None):  # noqa: D417
        """Determine optional prerequisite Nodes for a composite.

        Args:
            parent (Node): Compositor node to add these prerequisites under
            prereqs (Sequence): Strings (names), floats (wavelengths), or
                                DataQuerys to analyze.

        """
        prereq_nodes, unknown_datasets = self._create_prerequisite_subtrees(parent, prereqs, query)

        for prereq, unknowns in unknown_datasets.items():
            u_str = ", ".join([str(x) for x in unknowns])
            LOG.debug("Skipping optional %s: Unknown dataset %s",
                      str(prereq), u_str)

        return prereq_nodes

    def _create_prerequisite_subtrees(self, parent, prereqs, query=None):  # noqa: D417
        """Determine prerequisite Nodes for a composite.

        Args:
            parent (Node): Compositor node to add these prerequisites under
            prereqs (Sequence): Strings (names), floats (wavelengths),
                                DataQuerys or Nodes to analyze.

        """
        prereq_nodes = []
        unknown_datasets = dict()
        if not prereqs:
            # this composite has no required prerequisites
            prereq_nodes.append(self.empty_node)
            self.add_child(parent, self.empty_node)
            return prereq_nodes, unknown_datasets

        for prereq in prereqs:
            try:
                if isinstance(prereq, Node):
                    node = prereq
                else:
                    node = self._create_subtree_for_key(prereq, query=query)
            except MissingDependencies as unknown:
                unknown_datasets[prereq] = unknown.missing_dependencies

            else:
                prereq_nodes.append(node)
                self.add_child(parent, node)

        return prereq_nodes, unknown_datasets


class _DataIDContainer(dict):
    """Special dictionary object that can handle dict operations based on dataset name, wavelength, or DataID.

    Note: Internal dictionary keys are `DataID` objects.

    """

    def keys(self):
        """Give currently contained keys."""
        # sort keys so things are a little more deterministic (.keys() is not)
        return sorted(super(_DataIDContainer, self).keys())

    def get_key(self, match_key):
        """Get multiple fully-specified keys that match the provided query.

        Args:
            match_key (DataID): DataID or DataQuery of query parameters to use for
                                searching. Can also be a string representing the
                                dataset name or a number representing the dataset
                                wavelength.

        """
        return get_key(match_key, self.keys())

    def __getitem__(self, item):
        """Get item from container."""
        try:
            # short circuit - try to get the object without more work
            return super(_DataIDContainer, self).__getitem__(item)
        except KeyError:
            key = self.get_key(item)
            return super(_DataIDContainer, self).__getitem__(key)

    def __contains__(self, item):
        """Check if item exists in container."""
        try:
            key = self.get_key(item)
        except KeyError:
            return False
        return super(_DataIDContainer, self).__contains__(key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'msu_gsa_l1b' reader."""
import os
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler
from satpy.tests.utils import make_dataid

SOLCONST = "273.59"


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def _get_data(self, num_scans, num_cols):
        data = {
            "Data/resolution_1km/Solar_Zenith_Angle":
                xr.DataArray(
                    da.ones((num_scans*4, num_cols*4), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
            "Geolocation/resolution_1km/Latitude":
                xr.DataArray(
                    da.ones((num_scans*4, num_cols*4), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
            "Geolocation/resolution_1km/Longitude":
                xr.DataArray(
                    da.ones((num_scans*4, num_cols*4), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
            "Data/resolution_1km/Radiance_01":
                xr.DataArray(
                    da.ones((num_scans*4, num_cols*4), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999., "F_solar_constant": SOLCONST
                    },
                    dims=("x", "y")),
            "Data/resolution_4km/Solar_Zenith_Angle":
                xr.DataArray(
                    da.ones((num_scans, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
            "Geolocation/resolution_4km/Latitude":
                xr.DataArray(
                    da.ones((num_scans, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
            "Geolocation/resolution_4km/Longitude":
                xr.DataArray(
                    da.ones((num_scans, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
            "Data/resolution_4km/Brightness_Temperature_09":
                xr.DataArray(
                    da.ones((num_scans, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        "scale": 0.01, "offset": 0., "fill_value": -999.
                    },
                    dims=("x", "y")),
        }
        return data

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        num_scans = 20
        num_cols = 2048
        global_attrs = {
            "/attr/timestamp_without_timezone": "2022-01-13T12:45:00",
            "/attr/satellite_observation_point_height": "38500.0",
            "/attr/satellite_observation_point_latitude": "71.25",
            "/attr/satellite_observation_point_longitude": "21.44",
        }

        data = self._get_data(num_scans, num_cols)

        test_content = {}
        test_content.update(global_attrs)
        test_content.update(data)
        return test_content


class TestMSUGSABReader:
    """Test MSU GS/A L1B Reader."""

    yaml_file = "msu_gsa_l1b.yaml"

    def setup_method(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.core.loading import load_reader
        from satpy.readers.msu_gsa_l1b import MSUGSAFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(MSUGSAFileHandler, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

        filenames = ["ArcticaM1_202201131245.h5"]
        self.reader = load_reader(self.reader_configs)
        files = self.reader.select_files_from_pathnames(filenames)
        self.reader.create_filehandlers(files)

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_irbt(self):
        """Test retrieval in brightness temperature."""
        ds_ids = [make_dataid(name="C09", calibration="brightness_temperature")]
        res = self.reader.load(ds_ids)
        assert "C09" in res
        assert res["C09"].attrs["calibration"] == "brightness_temperature"
        assert res["C09"].attrs["platform_name"] == "Arctica-M-N1"
        assert res["C09"].attrs["sat_latitude"] == 71.25
        assert res["C09"].attrs["sat_longitude"] == 21.44
        assert res["C09"].attrs["sat_altitude"] == 38500.
        assert res["C09"].attrs["resolution"] == 4000

    def test_nocounts(self):
        """Test we can't get IR or VIS data as counts."""
        ds_ids = [make_dataid(name="C01", calibration="counts")]
        with pytest.raises(KeyError):
            self.reader.load(ds_ids)

        ds_ids = [make_dataid(name="C09", calibration="counts")]
        with pytest.raises(KeyError):
            self.reader.load(ds_ids)

    def test_vis_cal(self):
        """Test that we can retrieve VIS data as both radiance and reflectance."""
        ds_ids = [make_dataid(name="C01", calibration="radiance")]
        res = self.reader.load(ds_ids)
        rad = res["C01"].data
        ds_ids = [make_dataid(name="C01", calibration="reflectance")]
        res = self.reader.load(ds_ids)
        refl = res["C01"].data

        # Check the RAD->REFL conversion
        np.testing.assert_allclose(100 * np.pi * rad / float(SOLCONST), refl)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for VIIRS compositors."""

import datetime as dt

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.tests.utils import assert_maximum_dask_computes


class TestVIIRSComposites:
    """Test various VIIRS-specific composites."""

    @pytest.fixture
    def area(self):
        """Return fake area for use with DNB tests."""
        rows = 5
        cols = 10
        area = AreaDefinition(
            "test", "test", "test",
            {"proj": "eqc", "lon_0": 0.0,
             "lat_0": 0.0},
            cols, rows,
            (-20037508.34, -10018754.17, 20037508.34, 10018754.17))
        return area

    @pytest.fixture
    def dnb(self, area):
        """Return fake channel 1 data for DNB tests."""
        dnb = np.zeros(area.shape) + 0.25
        dnb[3, :] += 0.25
        dnb[4:, :] += 0.5
        dnb = da.from_array(dnb, chunks=2)
        c01 = xr.DataArray(dnb,
                           dims=("y", "x"),
                           attrs={"name": "DNB", "area": area,
                                  "start_time": dt.datetime(2020, 1, 1, 12, 0, 0)})
        return c01

    @pytest.fixture
    def sza(self, area):
        """Return fake sza dataset for DNB tests."""
        # data changes by row, sza changes by col for testing
        sza = np.zeros(area.shape) + 70.0
        sza[:, 3] += 20.0
        sza[:, 4:] += 45.0
        sza = da.from_array(sza, chunks=2)
        c02 = xr.DataArray(sza,
                           dims=("y", "x"),
                           attrs={"name": "solar_zenith_angle", "area": area,
                                  "start_time": dt.datetime(2020, 1, 1, 12, 0, 0)})
        return c02

    @pytest.fixture
    def lza(self, area):
        """Return fake lunal zenith angle dataset for DNB tests."""
        lza = np.zeros(area.shape) + 70.0
        lza[:, 3] += 20.0
        lza[:, 4:] += 45.0
        lza = da.from_array(lza, chunks=2)
        c03 = xr.DataArray(lza,
                           dims=("y", "x"),
                           attrs={"name": "lunar_zenith_angle", "area": area,
                                  "start_time": dt.datetime(2020, 1, 1, 12, 0, 0)
                                  })
        return c03

    def test_load_composite_yaml(self):
        """Test loading the yaml for this sensor."""
        from satpy.composites.config_loader import load_compositor_configs_for_sensors
        load_compositor_configs_for_sensors(["viirs"])

    def test_histogram_dnb(self, dnb, sza):
        """Test the 'histogram_dnb' compositor."""
        from satpy.composites.viirs import HistogramDNB

        with assert_maximum_dask_computes(max_computes=0):
            comp = HistogramDNB("histogram_dnb", prerequisites=("dnb",),
                                standard_name="toa_outgoing_radiance_per_"
                                              "unit_wavelength")
            res = comp((dnb, sza))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "histogram_dnb"
        assert res.attrs["standard_name"] == "equalized_radiance"
        data = res.compute()
        unique_values = np.unique(data)
        np.testing.assert_allclose(unique_values, [0.5994, 0.7992, 0.999], rtol=1e-3)

    def test_adaptive_dnb(self, dnb, sza):
        """Test the 'adaptive_dnb' compositor."""
        from satpy.composites.viirs import AdaptiveDNB

        with assert_maximum_dask_computes(max_computes=0):
            comp = AdaptiveDNB("adaptive_dnb", prerequisites=("dnb",),
                               standard_name="toa_outgoing_radiance_per_"
                                             "unit_wavelength")
            res = comp((dnb, sza))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "adaptive_dnb"
        assert res.attrs["standard_name"] == "equalized_radiance"
        data = res.compute()
        np.testing.assert_allclose(data.data, 0.999, rtol=1e-4)

    def test_hncc_dnb(self, area, dnb, sza, lza):
        """Test the 'hncc_dnb' compositor."""
        from satpy.composites.viirs import NCCZinke

        with assert_maximum_dask_computes(max_computes=0):
            comp = NCCZinke("hncc_dnb", prerequisites=("dnb",),
                            standard_name="toa_outgoing_radiance_per_"
                                          "unit_wavelength")
            mif = xr.DataArray(da.zeros((5,), chunks=5) + 0.1,
                               dims=("y",),
                               attrs={"name": "moon_illumination_fraction", "area": area})
            res = comp((dnb, sza, lza, mif))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "hncc_dnb"
        assert res.attrs["standard_name"] == "ncc_radiance"
        data = res.compute()
        unique = np.unique(data)
        np.testing.assert_allclose(
            unique, [3.48479712e-04, 6.96955799e-04, 1.04543189e-03, 4.75394738e-03,
                     9.50784532e-03, 1.42617433e-02, 1.50001560e+03, 3.00001560e+03,
                     4.50001560e+03])

        with pytest.raises(ValueError, match="Expected .*, got 2"):
            comp((dnb, sza))

    def test_hncc_dnb_nomoonpha(self, area, dnb, sza, lza):
        """Test the 'hncc_dnb' compositor when no moon phase data is provided."""
        from satpy.composites.viirs import NCCZinke

        with assert_maximum_dask_computes(max_computes=0):
            comp = NCCZinke("hncc_dnb", prerequisites=("dnb",),
                            standard_name="toa_outgoing_radiance_per_"
                                          "unit_wavelength")
            res = comp((dnb, sza, lza))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "hncc_dnb"
        assert res.attrs["standard_name"] == "ncc_radiance"
        data = res.compute()
        unique = np.unique(data)
        np.testing.assert_allclose(
            unique, [3.48479672e-04, 6.96955721e-04, 1.04543177e-03, 4.75394684e-03,
                     9.50784425e-03, 1.42617416e-02, 1.50001543e+03, 3.00001526e+03,
                     4.50001509e+03])

    @pytest.mark.parametrize("dnb_units", ["W m-2 sr-1", "W cm-2 sr-1"])
    @pytest.mark.parametrize("saturation_correction", [False, True])
    def test_erf_dnb(self, dnb_units, saturation_correction, area, sza, lza):
        """Test the 'dynamic_dnb' or ERF DNB compositor."""
        from satpy.composites.viirs import ERFDNB

        # dnb is different from in the other tests, so don't use the fixture
        # here
        dnb = np.zeros(area.shape) + 0.25
        cols = area.shape[1]
        dnb[2, :cols // 2] = np.nan
        dnb[3, :] += 0.25
        dnb[4:, :] += 0.5
        if dnb_units == "W cm-2 sr-1":
            dnb /= 10000.0
        dnb = da.from_array(dnb, chunks=2)
        c01 = xr.DataArray(dnb,
                           dims=("y", "x"),
                           attrs={"name": "DNB", "area": area, "units": dnb_units})
        mif = xr.DataArray(da.zeros((5,), chunks=5) + 0.1,
                           dims=("y",),
                           attrs={"name": "moon_illumination_fraction", "area": area})

        with assert_maximum_dask_computes(max_computes=0):
            comp = ERFDNB("dynamic_dnb", prerequisites=("dnb",),
                          saturation_correction=saturation_correction,
                          standard_name="toa_outgoing_radiance_per_"
                                        "unit_wavelength")
            res = comp((c01, sza, lza, mif))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "dynamic_dnb"
        assert res.attrs["standard_name"] == "equalized_radiance"
        data = res.compute()
        unique = np.unique(data)
        assert np.isnan(unique).any()
        nonnan_unique = unique[~np.isnan(unique)]
        if saturation_correction:
            exp_unique = [0.000000e+00, 3.978305e-04, 6.500003e-04,
                          8.286927e-04, 5.628335e-01, 7.959671e-01,
                          9.748567e-01]
        else:
            exp_unique = [0.00000000e+00, 1.00446703e-01, 1.64116082e-01,
                          2.09233451e-01, 1.43916324e+02, 2.03528498e+02,
                          2.49270516e+02]
        np.testing.assert_allclose(nonnan_unique, exp_unique)

    def test_snow_age(self, area):
        """Test the 'snow_age' compositor."""
        from satpy.composites.viirs import SnowAge

        projectables = tuple(
           xr.DataArray(
                da.from_array(np.full(area.shape, 5.*i), chunks=5),
                dims=("y", "x"),
                attrs={"name": f"M0{i:d}",
                       "calibration": "reflectance",
                       "units": "%"})
           for i in range(7, 12))
        with assert_maximum_dask_computes(max_computes=0):
            comp = SnowAge(
                    "snow_age",
                    prerequisites=("M07", "M08", "M09", "M10", "M11",),
                    standard_name="snow_age")
            res = comp(projectables)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "snow_age"
        assert "units" not in res.attrs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to VIIRS SDR format.

This reader implements the support of VIIRS SDR files as produced by CSPP and CLASS.
It is comprised of two parts:

 - A subclass of the YAMLFileReader class to allow handling all the files
 - A filehandler class to implement the actual reading

Format documentation:

 - http://npp.gsfc.nasa.gov/science/sciencedocuments/082012/474-00001-03_CDFCBVolIII_RevC.pdf

"""

import datetime as dt
import logging
import os.path
from contextlib import suppress
from glob import glob

import numpy as np

from satpy.readers.core.viirs_atms_sdr import ATMS_DATASET_KEYS, DATASET_KEYS, VIIRS_DATASET_KEYS, JPSS_SDR_FileHandler
from satpy.readers.core.yaml_reader import FileYAMLReader

NO_DATE = dt.datetime(1958, 1, 1)
EPSILON_TIME = dt.timedelta(days=2)
LOG = logging.getLogger(__name__)


def _get_invalid_info(granule_data):
    """Get a detailed report of the missing data.

    N/A: not applicable
    MISS: required value missing at time of processing
    OBPT: onboard pixel trim (overlapping/bow-tie pixel removed during SDR processing)
    OGPT: on-ground pixel trim (overlapping/bow-tie pixel removed during EDR processing)
    ERR: error occurred during processing / non-convergence
    ELINT: ellipsoid intersect failed / instrument line-of-sight does not intersect the Earth’s surface
    VDNE: value does not exist / processing algorithm did not execute
    SOUB: scaled out-of-bounds / solution not within allowed range
    """
    msg = None
    if issubclass(granule_data.dtype.type, np.integer):
        msg = ("na:" + str((granule_data == 65535).sum()) +
               " miss:" + str((granule_data == 65534).sum()) +
               " obpt:" + str((granule_data == 65533).sum()) +
               " ogpt:" + str((granule_data == 65532).sum()) +
               " err:" + str((granule_data == 65531).sum()) +
               " elint:" + str((granule_data == 65530).sum()) +
               " vdne:" + str((granule_data == 65529).sum()) +
               " soub:" + str((granule_data == 65528).sum()))
    elif issubclass(granule_data.dtype.type, np.floating):
        msg = ("na:" + str((granule_data == -999.9).sum()) +
               " miss:" + str((granule_data == -999.8).sum()) +
               " obpt:" + str((granule_data == -999.7).sum()) +
               " ogpt:" + str((granule_data == -999.6).sum()) +
               " err:" + str((granule_data == -999.5).sum()) +
               " elint:" + str((granule_data == -999.4).sum()) +
               " vdne:" + str((granule_data == -999.3).sum()) +
               " soub:" + str((granule_data == -999.2).sum()))
    return msg


class VIIRSSDRFileHandler(JPSS_SDR_FileHandler):
    """VIIRS SDR HDF5 File Reader."""

    def __init__(self, filename, filename_info, filetype_info, use_tc=None, **kwargs):
        """Initialize file handler."""
        self.datasets = filename_info["datasets"].split("-")
        self.use_tc = use_tc
        super().__init__(filename, filename_info, filetype_info, **kwargs)

    def __getitem__(self, item):
        """Get item."""
        if "*" in item:
            # this is an aggregated field that can't easily be loaded, need to
            # join things together
            idx = 0
            base_item = item
            item = base_item.replace("*", str(idx))
            result = []
            while True:
                try:
                    res = super().__getitem__(item)
                    result.append(res)
                except KeyError:
                    # no more granule keys
                    LOG.debug("Aggregated granule stopping on '%s'", item)
                    break

                idx += 1
                item = base_item.replace("*", str(idx))
            return result
        else:
            return super().__getitem__(item)

    def get_dataset(self, dataset_id, ds_info):
        """Get the dataset corresponding to *dataset_id*.

        The size of the return DataArray will be dependent on the number of
        scans actually sensed, and not necessarily the regular 768 scanlines
        that the file contains for each granule. To that end, the number of
        scans for each granule is read from:
        ``Data_Products/...Gran_x/N_Number_Of_Scans``.
        """
        dataset_group = [ds_group for ds_group in ds_info["dataset_groups"] if ds_group in self.datasets]
        if not dataset_group:
            return
        dataset_group = dataset_group[0]
        ds_info["dataset_group"] = dataset_group
        var_path = self._generate_file_key(dataset_id, ds_info)

        data = self.concatenate_dataset(dataset_group, var_path)
        data = self.mask_fill_values(data, ds_info)

        data = self.scale_data_to_specified_unit(data, dataset_id, ds_info)
        data = self._update_data_attributes(data, dataset_id, ds_info)

        return data

    def get_bounding_box(self):
        """Get the bounding box of this file."""
        from pyproj import Geod
        geod = Geod(ellps="WGS84")
        dataset_group = DATASET_KEYS[self.datasets[0]]
        idx = 0
        lons_ring = None
        lats_ring = None
        while True:
            path = "Data_Products/{dataset_group}/{dataset_group}_Gran_{idx}/attr/"
            prefix = path.format(dataset_group=dataset_group, idx=idx)
            try:
                lats = self.file_content[prefix + "G-Ring_Latitude"]
                lons = self.file_content[prefix + "G-Ring_Longitude"]
                if lons_ring is None:
                    lons_ring = lons
                    lats_ring = lats
                else:
                    prev_lon = lons_ring[0]
                    prev_lat = lats_ring[0]
                    dists = [geod.inv(lon, lat, prev_lon, prev_lat)[2] for lon, lat in zip(lons, lats)]
                    first_idx = np.argmin(dists)
                    if first_idx == 2 and len(lons) == 8:
                        lons_ring = np.hstack((lons[:3], lons_ring[:-2], lons[4:]))
                        lats_ring = np.hstack((lats[:3], lats_ring[:-2], lats[4:]))
                    else:
                        raise NotImplementedError("Don't know how to handle G-Rings of length %d" % len(lons))

            except KeyError:
                break
            idx += 1

        return lons_ring, lats_ring


def split_desired_other(fhs, prime_geo, second_geo):
    """Split the provided filehandlers *fhs* into desired filehandlers and others."""
    desired = []
    other = []
    for fh in fhs:
        if prime_geo in fh.datasets:
            desired.append(fh)
        elif second_geo in fh.datasets:
            other.append(fh)
    return desired, other


class VIIRSSDRReader(FileYAMLReader):
    """Custom file reader for finding VIIRS SDR geolocation at runtime."""

    def __init__(self, config_files, use_tc=None, **kwargs):  # noqa: D417
        """Initialize file reader and adjust geolocation preferences.

        Args:
            config_files (Iterable): yaml config files passed to base class
            use_tc (bool): If `True` use the terrain corrected
                           files. If `False`, switch to non-TC files. If
                           `None` (default), use TC if available, non-TC otherwise.

        """
        super().__init__(config_files, **kwargs)
        self.use_tc = use_tc

    def _is_viirs_dataset(self, datasets):
        for dataset in VIIRS_DATASET_KEYS:
            if dataset in datasets:
                return True
        return False

    def filter_filenames_by_info(self, filename_items):
        """Filter out file using metadata from the filenames.

        This sorts out the different lon and lat datasets depending on TC is
        desired or not.
        """
        filename_items = list(filename_items)
        geo_keep = []
        geo_del = []
        viirs_del = []
        for filename, filename_info in filename_items:
            datasets = filename_info["datasets"].split("-")
            if not self._is_viirs_dataset(datasets):
                viirs_del.append(filename)

            if ("GITCO" in datasets) or ("GMTCO" in datasets):
                if self.use_tc is False:
                    geo_del.append(filename)
                else:
                    geo_keep.append(filename)
            elif ("GIMGO" in datasets) or ("GMODO" in datasets):
                if self.use_tc is True:
                    geo_del.append(filename)
                else:
                    geo_keep.append(filename)
        if geo_keep:
            filename_items = self._remove_geo_datasets_from_files(filename_items, geo_del)

        filename_items = self._remove_non_viirs_datasets_from_files(filename_items, viirs_del)
        return super().filter_filenames_by_info(filename_items)

    def _remove_non_viirs_datasets_from_files(self, filename_items, files_to_edit):
        no_viirs = ATMS_DATASET_KEYS
        return self._remove_datasets_from_files(filename_items, files_to_edit, no_viirs)

    def _remove_geo_datasets_from_files(self, filename_items, files_to_edit):
        datasets_to_consider = ["GITCO", "GMTCO", "GIMGO", "GMODO"]
        return self._remove_datasets_from_files(filename_items, files_to_edit, datasets_to_consider)

    def _remove_datasets_from_files(self, filename_items, files_to_edit, considered_datasets):
        fdict = dict(filename_items)
        for to_del in files_to_edit:
            fdict[to_del]["datasets"] = fdict[to_del]["datasets"].split("-")
            for dataset in considered_datasets:
                with suppress(ValueError):
                    fdict[to_del]["datasets"].remove(dataset)
            if not fdict[to_del]["datasets"]:
                del fdict[to_del]
            else:
                fdict[to_del]["datasets"] = "-".join(fdict[to_del]["datasets"])
        filename_items = fdict.items()
        return filename_items

    def _load_filenames_from_geo_ref(self, dsid):
        """Load filenames from the N_GEO_Ref attribute of a dataset's file."""
        file_handlers = self._get_file_handlers(dsid)
        if not file_handlers:
            return []

        fns = []
        for fh in file_handlers:
            base_dir = os.path.dirname(fh.filename)
            try:
                # get the filename and remove the creation time
                # which is often wrong
                fn = fh["/attr/N_GEO_Ref"][:46] + "*.h5"
                fns.extend(glob(os.path.join(base_dir, fn)))

                # usually is non-terrain corrected file, add the terrain
                # corrected file too
                if fn[:5] == "GIMGO":
                    fn = "GITCO" + fn[5:]
                elif fn[:5] == "GMODO":
                    fn = "GMTCO" + fn[5:]
                else:
                    continue
                fns.extend(glob(os.path.join(base_dir, fn)))
            except KeyError:
                LOG.debug("Could not load geo-reference information from {}".format(fh.filename))

        return fns

    def _get_primary_secondary_geo_groups(self, ds_info):
        """Find out which geolocation files are needed."""
        if ds_info["dataset_groups"][0].startswith("GM"):
            if self.use_tc is False:
                prime_geo = "GMODO"
                second_geo = "GMTCO"
            else:
                prime_geo = "GMTCO"
                second_geo = "GMODO"
        elif ds_info["dataset_groups"][0].startswith("GI"):
            if self.use_tc is False:
                prime_geo = "GIMGO"
                second_geo = "GITCO"
            else:
                prime_geo = "GITCO"
                second_geo = "GIMGO"
        else:
            raise ValueError("Unknown dataset group %s" % ds_info["dataset_groups"][0])
        return prime_geo, second_geo

    def get_right_geo_fhs(self, dsid, fhs):
        """Find the right geographical file handlers for given dataset ID *dsid*."""
        ds_info = self.all_ids[dsid]
        prime_geo, second_geo = self._get_primary_secondary_geo_groups(ds_info)
        desired, other = split_desired_other(fhs, prime_geo, second_geo)
        if desired:
            try:
                ds_info["dataset_groups"].remove(second_geo)
            except ValueError:
                pass
            return desired
        else:
            return other

    def _get_file_handlers(self, dsid):
        """Get the file handler to load this dataset."""
        ds_info = self.all_ids[dsid]

        fhs = [fh for fh in self.file_handlers["generic_file"]
               if set(fh.datasets) & set(ds_info["dataset_groups"])]
        if not fhs:
            LOG.warning("Required file type '%s' not found or loaded for "
                        "'%s'", ds_info["file_type"], dsid["name"])
        else:
            if len(set(ds_info["dataset_groups"]) & {"GITCO", "GIMGO", "GMTCO", "GMODO"}) > 1:
                fhs = self.get_right_geo_fhs(dsid, fhs)
            return fhs

    def _get_coordinates_for_dataset_key(self, dsid):
        """Get the coordinate dataset keys for `dsid`.

        Wraps the base class method in order to load geolocation files
        from the geo reference attribute in the datasets file.
        """
        coords = super()._get_coordinates_for_dataset_key(dsid)
        for c_id in coords:
            c_info = self.all_ids[c_id]  # c_info['dataset_groups'] should be a list of 2 elements
            self._get_file_handlers(c_id)
            prime_geo, second_geo = self._geo_dataset_groups(c_info)
            if prime_geo is None:
                continue

            # check the dataset file for the geolocation filename
            geo_filenames = self._load_filenames_from_geo_ref(dsid)
            self._create_new_geo_file_handlers(geo_filenames)
            self._remove_not_loaded_geo_dataset_group(c_info["dataset_groups"], prime_geo, second_geo)

        return coords

    def _geo_dataset_groups(self, c_info):
        if len(c_info["dataset_groups"]) == 1:  # filtering already done
            return None, None
        try:
            prime_geo, second_geo = self._get_primary_secondary_geo_groups(c_info)
            return prime_geo, second_geo
        except ValueError:  # DNB
            return None, None

    def _create_new_geo_file_handlers(self, geo_filenames):
        existing_filenames = set([fh.filename for fh in self.file_handlers["generic_file"]])
        geo_filenames = set(geo_filenames) - existing_filenames
        self.create_filehandlers(geo_filenames)

    def _remove_not_loaded_geo_dataset_group(self, c_dataset_groups, prime_geo, second_geo):
        all_fhs = self.file_handlers["generic_file"]
        desired, other = split_desired_other(all_fhs, prime_geo, second_geo)
        group_to_remove = second_geo if desired else prime_geo
        c_dataset_groups.remove(group_to_remove)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""CF decoding."""

import copy
import datetime as dt
import json


def decode_attrs(attrs):
    """Decode CF-encoded attributes to Python object.

    Converts timestamps to datetime and strings starting with "{" to
    dictionary.

    Args:
        attrs (dict): Attributes to be decoded

    Returns (dict): Decoded attributes
    """
    attrs = copy.deepcopy(attrs)
    _decode_dict_type_attrs(attrs)
    _decode_timestamps(attrs)
    return attrs


def _decode_dict_type_attrs(attrs):
    for key, val in attrs.items():
        attrs[key] = _str2dict(val)


def _str2dict(val):
    """Convert string to dictionary."""
    if isinstance(val, str) and val.startswith("{"):
        val = json.loads(val, object_hook=_datetime_parser_json)
    return val


def _decode_timestamps(attrs):
    for key, value in attrs.items():
        timestamp = _str2datetime(value)
        if timestamp:
            attrs[key] = timestamp


def _datetime_parser_json(json_dict):
    """Traverse JSON dictionary and parse timestamps."""
    for key, value in json_dict.items():
        timestamp = _str2datetime(value)
        if timestamp:
            json_dict[key] = timestamp
    return json_dict


def _str2datetime(string):
    """Convert string to datetime object."""
    try:
        return dt.datetime.fromisoformat(string)
    except (TypeError, ValueError):
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Sentinel-3 OLCI reader.

This reader supports an optional argument to choose the 'engine' for reading
OLCI netCDF4 files. By default, this reader uses the default xarray choice of
engine, as defined in the :func:`xarray.open_dataset` documentation`.

As an alternative, the user may wish to use the 'h5netcdf' engine, but that is
not default as it typically prints many non-fatal but confusing error messages
to the terminal.
To choose between engines the user can  do as follows for the default::

    scn = Scene(filenames=my_files, reader='olci_l1b')

or as follows for the h5netcdf engine::

    scn = Scene(filenames=my_files,
                reader='olci_l1b', reader_kwargs={'engine': 'h5netcdf'})

References:
    - :func:`xarray.open_dataset`

"""


import logging
from functools import reduce

import dask.array as da
import numpy as np
import xarray as xr

from satpy._compat import cached_property
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.utils import angle2xyz, get_legacy_chunk_size, xyz2angle

# the order of the L1B quality flags are from highest 32nd bit to the lowest 1 bit
# https://sentinel.esa.int/documents/247904/1872756/Sentinel-3-OLCI-Product-Data-Format-Specification-OLCI-Level-1
L1B_QUALITY_FLAGS = ["saturated@Oa21", "saturated@Oa20", "saturated@Oa19", "saturated@Oa18",
                     "saturated@Oa17", "saturated@Oa16", "saturated@Oa15", "saturated@Oa14",
                     "saturated@Oa13", "saturated@Oa12", "saturated@Oa11", "saturated@Oa10",
                     "saturated@Oa09", "saturated@Oa08", "saturated@Oa07", "saturated@Oa06",
                     "saturated@Oa05", "saturated@Oa04", "saturated@Oa03", "saturated@Oa02",
                     "saturated@Oa01", "dubious", "sun-glint_risk", "duplicated",
                     "cosmetic", "invalid", "straylight_risk", "bright",
                     "tidal_region", "fresh_inland_water", "coastline", "land"]

DEFAULT_L1B_MASK_ITEMS = ["dubious", "sun-glint_risk", "duplicated", "cosmetic", "invalid",
    "straylight_risk", "bright", "tidal_region", "coastline", "land"]

WQSF_FLAG_LIST = ["INVALID", "WATER", "LAND", "CLOUD", "SNOW_ICE", "INLAND_WATER", "TIDAL",
    "COSMETIC", "SUSPECT", "HISOLZEN", "SATURATED", "MEGLINT", "HIGHGLINT", "WHITECAPS", "ADJAC",
    "WV_FAIL", "PAR_FAIL", "AC_FAIL", "OC4ME_FAIL", "OCNN_FAIL", "Extra_1", "KDM_FAIL", "Extra_2",
    "CLOUD_AMBIGUOUS", "CLOUD_MARGIN", "BPAC_ON", "WHITE_SCATT", "LOWRW", "HIGHRW"]

DEFAULT_WQSF_MASK_ITEMS = ["INVALID", "SNOW_ICE", "INLAND_WATER", "SUSPECT", "AC_FAIL", "CLOUD",
    "HISOLZEN", "OCNN_FAIL", "CLOUD_MARGIN", "CLOUD_AMBIGUOUS", "LOWRW", "LAND"]

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

PLATFORM_NAMES = {"S3A": "Sentinel-3A",
                  "S3B": "Sentinel-3B",
                  "ENV": "Environmental Satellite"}


class BitFlags:
    """Manipulate flags stored bitwise."""

    def __init__(self, value, flag_list=None):
        """Init the flags."""
        self._value = value

        if flag_list is None:
            try:
                meanings = value.attrs["flag_meanings"].split()
                masks = value.attrs["flag_masks"]
            except (AttributeError, KeyError):
                meanings = WQSF_FLAG_LIST
                self.meaning = {meaning: mask for mask, meaning in enumerate(meanings)}
            else:
                self.meaning = {meaning: int(np.log2(mask)) for meaning, mask in zip(meanings, masks)}
        else:
            self.meaning = {meaning: mask for mask, meaning in enumerate(flag_list)}

    def __getitem__(self, item):
        """Get the item."""
        pos = self.meaning[item]
        data = self._value
        if isinstance(data, xr.DataArray):
            data = data.data
            res = ((data >> pos) % 2).astype(bool)
            res = xr.DataArray(res, coords=self._value.coords, attrs=self._value.attrs,
                               dims=self._value.dims)
        else:
            res = ((data >> pos) % 2).astype(bool)
        return res


class NCOLCIBase(BaseFileHandler):
    """The OLCI reader base."""

    rows_name = "rows"
    cols_name = "columns"

    def __init__(self, filename, filename_info, filetype_info, engine=None, **kwargs):
        """Init the olci reader base."""
        super().__init__(filename, filename_info, filetype_info)
        self._engine = engine
        self._start_time = filename_info["start_time"]
        self._end_time = filename_info["end_time"]
        # TODO: get metadata from the manifest file (xfdumanifest.xml)
        self.platform_name = PLATFORM_NAMES[filename_info["mission_id"]]
        self.sensor = "olci"

    @cached_property
    def nc(self):
        """Get the nc xr dataset."""
        f_obj = open_file_or_filename(self.filename)
        dataset = xr.open_dataset(f_obj,
                                  decode_cf=True,
                                  mask_and_scale=True,
                                  engine=self._engine,
                                  chunks={self.cols_name: CHUNK_SIZE,
                                          self.rows_name: CHUNK_SIZE})
        return dataset.rename({self.cols_name: "x", self.rows_name: "y"})

    @property
    def start_time(self):
        """Start time property."""
        return self._start_time

    @property
    def end_time(self):
        """End time property."""
        return self._end_time

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading %s.", key["name"])
        variable = self.nc[key["name"]]

        return variable


class NCOLCICal(NCOLCIBase):
    """Dummy class for calibration."""


class NCOLCIGeo(NCOLCIBase):
    """Dummy class for navigation."""


class NCOLCIChannelBase(NCOLCIBase):
    """Base class for channel reading."""

    def __init__(self, filename, filename_info, filetype_info, engine=None):
        """Init the file handler."""
        super().__init__(filename, filename_info, filetype_info, engine)
        self.channel = filename_info.get("dataset_name")
        self.reflectance_prefix = "Oa"
        self.reflectance_suffix = "_reflectance"


class NCOLCI1B(NCOLCIChannelBase):
    """File handler for OLCI l1b."""

    def __init__(self, filename, filename_info, filetype_info, cal=None, engine=None, mask_items=None):
        """Init the file handler."""
        super().__init__(filename, filename_info, filetype_info, engine)
        if cal is not None:
            self.cal = cal.nc
        self.mask_items = mask_items

    @staticmethod
    def _get_items(idx, solar_flux):
        """Get items."""
        return solar_flux[idx]

    def _get_solar_flux(self, band):
        """Get the solar flux for the band."""
        solar_flux = self.cal["solar_flux"].isel(bands=band).values
        d_index = self.cal["detector_index"].fillna(0).astype(int)

        return da.map_blocks(self._get_items, d_index.data,
                             solar_flux=solar_flux, dtype=solar_flux.dtype)

    def getbitmask(self, quality_flags, items=None):
        """Get the quality flags bitmask."""
        if items is None:
            items = DEFAULT_L1B_MASK_ITEMS

        bflags = BitFlags(quality_flags, flag_list=L1B_QUALITY_FLAGS)

        return reduce(np.logical_or, [bflags[item] for item in items])

    def get_dataset(self, key, info):
        """Load a dataset."""
        if self.channel is not None and self.channel != key["name"]:
            return

        logger.debug("Reading %s.", key["name"])

        if key["name"] == "quality_flags":
            dataset = self.nc["quality_flags"]
        elif key["name"] == "mask":
            dataset = self.getbitmask(self.nc["quality_flags"], self.mask_items)
        else:
            dataset = self.nc[self.channel + "_radiance"]

            if key["calibration"] == "reflectance":
                idx = int(key["name"][2:]) - 1
                sflux = self._get_solar_flux(idx)
                dataset = dataset / sflux * np.pi * 100
                dataset.attrs["units"] = "%"

        dataset.attrs["platform_name"] = self.platform_name
        dataset.attrs["sensor"] = self.sensor
        dataset.attrs.update(key.to_dict())

        return dataset


class NCOLCI2(NCOLCIChannelBase):
    """File handler for OLCI l2."""

    def __init__(self, filename, filename_info, filetype_info, engine=None, unlog=False, mask_items=None):
        """Init the file handler."""
        super().__init__(filename, filename_info, filetype_info, engine)
        self.unlog = unlog
        self.mask_items = mask_items

    def get_dataset(self, key, info):
        """Load a dataset."""
        if self.channel is not None and self.channel != key["name"]:
            return
        logger.debug("Reading %s.", key["name"])

        if self.channel is not None and self.channel.startswith(self.reflectance_prefix):
            dataset = self.nc[self.channel + self.reflectance_suffix]
        else:
            dataset = self.nc[info["nc_key"]]

        if key["name"] == "wqsf":
            dataset.attrs["_FillValue"] = 1
        elif key["name"] == "mask":
            dataset = self.getbitmask(dataset, self.mask_items)

        dataset.attrs["platform_name"] = self.platform_name
        dataset.attrs["sensor"] = self.sensor
        dataset.attrs.update(key.to_dict())
        if self.unlog:
            dataset = self.delog(dataset)

        return dataset

    def delog(self, data_array):
        """Remove log10 from the units and values."""
        units = data_array.attrs["units"]

        if units.startswith("lg("):
            data_array = 10 ** data_array
            data_array.attrs["units"] = units.split("lg(re ")[1].strip(")")
        return data_array

    def getbitmask(self, wqsf, items=None):
        """Get the bitmask."""
        if items is None:
            items = DEFAULT_WQSF_MASK_ITEMS
        bflags = BitFlags(wqsf, WQSF_FLAG_LIST)
        return reduce(np.logical_or, [bflags[item] for item in items])


class NCOLCILowResData(NCOLCIBase):
    """Handler for low resolution data."""

    rows_name = "tie_rows"
    cols_name = "tie_columns"

    def __init__(self, filename, filename_info, filetype_info,
                 engine=None, **kwargs):
        """Init the file handler."""
        super().__init__(filename, filename_info, filetype_info, engine)

    @property
    def l_step(self):
        """Get the line step."""
        return self.nc.attrs["al_subsampling_factor"]

    @property
    def c_step(self):
        """Get the column step."""
        return self.nc.attrs["ac_subsampling_factor"]

    def _do_interpolate(self, data):

        if not isinstance(data, tuple):
            data = (data,)

        shape = data[0].shape

        from geotiepoints.interpolator import Interpolator
        tie_lines = np.arange(0, (shape[0] - 1) * self.l_step + 1, self.l_step)
        tie_cols = np.arange(0, (shape[1] - 1) * self.c_step + 1, self.c_step)
        lines = np.arange((shape[0] - 1) * self.l_step + 1)
        cols = np.arange((shape[1] - 1) * self.c_step + 1)
        along_track_order = 1
        cross_track_order = 3
        satint = Interpolator([x.values for x in data],
                              (tie_lines, tie_cols),
                              (lines, cols),
                              along_track_order,
                              cross_track_order)
        int_data = satint.interpolate()

        return [xr.DataArray(da.from_array(x, chunks=(CHUNK_SIZE, CHUNK_SIZE)),
                             dims=["y", "x"]) for x in int_data]

    @property
    def _need_interpolation(self):
        return (self.c_step != 1 or self.l_step != 1)


class NCOLCIAngles(NCOLCILowResData):
    """File handler for the OLCI angles."""

    datasets = {"satellite_azimuth_angle": "OAA",
                "satellite_zenith_angle": "OZA",
                "solar_azimuth_angle": "SAA",
                "solar_zenith_angle": "SZA"}

    def get_dataset(self, key, info):
        """Load a dataset."""
        if key["name"] not in self.datasets:
            return

        logger.debug("Reading %s.", key["name"])

        if self._need_interpolation:
            if key["name"].startswith("satellite"):
                azi, zen = self.satellite_angles
            elif key["name"].startswith("solar"):
                azi, zen = self.sun_angles
            else:
                raise NotImplementedError("Don't know how to read " + key["name"])

            if "zenith" in key["name"]:
                values = zen
            elif "azimuth" in key["name"]:
                values = azi
            else:
                raise NotImplementedError("Don't know how to read " + key["name"])
        else:
            values = self.nc[self.datasets[key["name"]]]

        values.attrs["platform_name"] = self.platform_name
        values.attrs["sensor"] = self.sensor

        values.attrs.update(key.to_dict())
        return values

    @cached_property
    def sun_angles(self):
        """Return the sun angles."""
        zen = self.nc[self.datasets["solar_zenith_angle"]]
        azi = self.nc[self.datasets["solar_azimuth_angle"]]
        azi, zen = self._interpolate_angles(azi, zen)
        return azi, zen

    @cached_property
    def satellite_angles(self):
        """Return the satellite angles."""
        zen = self.nc[self.datasets["satellite_zenith_angle"]]
        azi = self.nc[self.datasets["satellite_azimuth_angle"]]
        azi, zen = self._interpolate_angles(azi, zen)
        return azi, zen

    def _interpolate_angles(self, azi, zen):
        aattrs = azi.attrs
        zattrs = zen.attrs
        x, y, z = angle2xyz(azi, zen)
        x, y, z = self._do_interpolate((x, y, z))
        azi, zen = xyz2angle(x, y, z)
        azi.attrs = aattrs
        zen.attrs = zattrs
        return azi, zen


class NCOLCIMeteo(NCOLCILowResData):
    """File handler for the OLCI meteo data."""

    datasets = ["humidity", "sea_level_pressure", "total_columnar_water_vapour", "total_ozone"]

    def __init__(self, filename, filename_info, filetype_info,
                 engine=None):
        """Init the file handler."""
        super().__init__(filename, filename_info, filetype_info, engine)
        self.cache = {}

    # TODO: the following depends on more than columns, rows
    # float atmospheric_temperature_profile(tie_rows, tie_columns, tie_pressure_levels) ;
    # float horizontal_wind(tie_rows, tie_columns, wind_vectors) ;
    # float reference_pressure_level(tie_pressure_levels) ;

    def get_dataset(self, key, info):
        """Load a dataset."""
        if key["name"] not in self.datasets:
            return

        logger.debug("Reading %s.", key["name"])

        if self._need_interpolation and self.cache.get(key["name"]) is None:

            data = self.nc[key["name"]]

            values, = self._do_interpolate(data)
            values.attrs = data.attrs

            self.cache[key["name"]] = values

        elif key["name"] in self.cache:
            values = self.cache[key["name"]]
        else:
            values = self.nc[key["name"]]

        values.attrs["platform_name"] = self.platform_name
        values.attrs["sensor"] = self.sensor

        values.attrs.update(key.to_dict())
        return values



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Core functionality of composites."""

from __future__ import annotations

import logging
import warnings
from typing import Optional, Sequence

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset import DataID, combine_metadata
from satpy.dataset.dataid import minimal_default_keys_config
from satpy.utils import unify_chunks

LOG = logging.getLogger(__name__)

NEGLIGIBLE_COORDS = ["time"]
"""Keywords identifying non-dimensional coordinates to be ignored during composite generation."""

TIME_COMPATIBILITY_TOLERANCE = np.timedelta64(1, "s")


class IncompatibleAreas(Exception):
    """Error raised upon compositing things of different shapes."""


class IncompatibleTimes(Exception):
    """Error raised upon compositing things from different times."""


class CompositeBase:
    """Base class for all compositors and modifiers.

    A compositor in Satpy is a class that takes in zero or more input
    DataArrays and produces a new DataArray with its own identifier (name).
    The result of a compositor is typically a brand new "product" that
    represents something different than the inputs that went into the
    operation.

    See the :class:`~satpy.modifiers.base.ModifierBase` class for information
    on the similar concept of "modifiers".

    """

    def __init__(self, name, prerequisites=None, optional_prerequisites=None, **kwargs):
        """Initialise the compositor."""
        # Required info
        kwargs["name"] = name
        kwargs["prerequisites"] = prerequisites or []
        kwargs["optional_prerequisites"] = optional_prerequisites or []
        self.attrs = kwargs

    @property
    def id(self):  # noqa: A003
        """Return the DataID of the object."""
        try:
            return self.attrs["_satpy_id"]
        except KeyError:
            id_keys = self.attrs.get("_satpy_id_keys", minimal_default_keys_config)
            return DataID(id_keys, **self.attrs)

    def __call__(
            self,
            datasets: Sequence[xr.DataArray],
            optional_datasets: Optional[Sequence[xr.DataArray]] = None,
            **info
    ) -> xr.DataArray:
        """Generate a composite."""
        raise NotImplementedError()

    def __str__(self):
        """Stringify the object."""
        from pprint import pformat
        return pformat(self.attrs)

    def __repr__(self):
        """Represent the object."""
        from pprint import pformat
        return pformat(self.attrs)

    def apply_modifier_info(self, origin, destination):
        """Apply the modifier info from *origin* to *destination*."""
        try:
            dataset_keys = self.attrs["_satpy_id"].id_keys.keys()
        except KeyError:
            dataset_keys = ["name", "modifiers"]

        self._collect_modifier_info(origin, destination, dataset_keys)

    def _collect_modifier_info(self, origin, destination, dataset_keys):
        o = getattr(origin, "attrs", origin)
        d = getattr(destination, "attrs", destination)

        for k in dataset_keys:
            if self._is_existing_modifier(k):
                d[k] = self.attrs[k]
            elif d.get(k) is None:
                self._add_missing_modifier(k, d, o)

    def _is_existing_modifier(self, k):
        return (k == "modifiers") and (k in self.attrs)

    def _add_missing_modifier(self, key, destination, origin):
        if self.attrs.get(key) is not None:
            destination[key] = self.attrs[key]
        elif origin.get(key) is not None:
            destination[key] = origin[key]

    def match_data_arrays(self, data_arrays: Sequence[xr.DataArray]) -> list[xr.DataArray]:
        """Match data arrays so that they can be used together in a composite.

        For the purpose of this method, "can be used together" means:

        - All arrays should have the same dimensions.
        - Either all arrays should have an area, or none should.
        - If all have an area, the areas should be all the same.

        In addition, negligible non-dimensional coordinates are dropped (see
        :meth:`drop_coordinates`) and dask chunks are unified (see
        :func:`satpy.utils.unify_chunks`).

        Args:
            data_arrays: Arrays to be checked

        Returns:
            Arrays with negligible non-dimensional coordinates removed.

        Raises:
            :class:`IncompatibleAreas`:
                If dimension or areas do not match.
            :class:`ValueError`:
                If some, but not all data arrays lack an area attribute.
        """
        self.check_geolocation(data_arrays)
        new_arrays = self.drop_coordinates(data_arrays)
        new_arrays = self.align_geo_coordinates(new_arrays)
        new_arrays = list(unify_chunks(*new_arrays))
        return new_arrays

    def check_geolocation(self, data_arrays: Sequence[xr.DataArray]) -> None:
        """Check that the geolocations of the *data_arrays* are compatible.

        For the purpose of this method, "compatible" means:

        - All arrays should have the same dimensions.
        - Either all arrays should have an area, or none should.
        - If all have an area, the areas should be all the same.

        Args:
            data_arrays: Arrays to be checked

        Raises:
            :class:`IncompatibleAreas`:
                If dimension or areas do not match.
            :class:`ValueError`:
                If some, but not all data arrays lack an area attribute.
        """
        if len(data_arrays) == 1:
            return

        self._check_dimension_size(data_arrays, "x")
        self._check_dimension_size(data_arrays, "y")

        areas = [ds.attrs.get("area") for ds in data_arrays]
        if all(a is None for a in areas):
            return

        self._check_areas_are_valid(areas)

    @staticmethod
    def _check_dimension_size(data_arrays, coordinate):
        if coordinate in data_arrays[0].dims and \
           not all(x.sizes[coordinate] == data_arrays[0].sizes[coordinate]
                   for x in data_arrays[1:]):
            coordinate = coordinate.upper()
            raise IncompatibleAreas(f"{coordinate} dimension has different sizes")

    def _check_areas_are_valid(self, areas):
        if any(a is None for a in areas):
            raise ValueError("Missing 'area' attribute")

        if not all(areas[0] == x for x in areas[1:]):
            LOG.debug("Not all areas are the same in "
                      "'{}'".format(self.attrs["name"]))
            raise IncompatibleAreas("Areas are different")

    @staticmethod
    def drop_coordinates(data_arrays: Sequence[xr.DataArray]) -> list[xr.DataArray]:
        """Drop negligible non-dimensional coordinates.

        Drops negligible coordinates if they do not correspond to any
        dimension.  Negligible coordinates are defined in the
        :attr:`NEGLIGIBLE_COORDS` module attribute.

        Args:
            data_arrays: Arrays to be checked
        """
        new_arrays = []
        for ds in data_arrays:
            drop = [coord for coord in ds.coords
                    if coord not in ds.dims and
                    any([neglible in coord for neglible in NEGLIGIBLE_COORDS])]
            if drop:
                new_arrays.append(ds.drop_vars(drop))
            else:
                new_arrays.append(ds)

        return new_arrays

    @staticmethod
    def align_geo_coordinates(data_arrays: Sequence[xr.DataArray]) -> list[xr.DataArray]:
        """Align DataArrays along geolocation coordinates.

        See :func:`~xarray.align` for more information. This function uses
        the "override" join method to essentially ignore differences between
        coordinates. The :meth:`check_geolocation` should be called before
        this to ensure that geolocation coordinates and "area" are compatible.
        The :meth:`drop_coordinates` method should be called before this to
        ensure that coordinates that are considered "negligible" when computing
        composites do not affect alignment.

        """
        non_geo_coords = tuple(
            coord_name for data_arr in data_arrays
            for coord_name in data_arr.coords if coord_name not in ("x", "y"))
        return list(xr.align(*data_arrays, join="override", exclude=non_geo_coords))


def enhance2dataset(dset, convert_p=False):
    """Return the enhancement dataset *dset* as an array.

    If `convert_p` is True, enhancements generating a P mode will be converted to RGB or RGBA.
    """
    attrs = dset.attrs
    data = _get_data_from_enhanced_image(dset, convert_p)
    data.attrs = attrs
    # remove 'mode' if it is specified since it may have been updated
    data.attrs.pop("mode", None)
    # update mode since it may have changed (colorized/palettize)
    data.attrs["mode"] = GenericCompositor.infer_mode(data)
    return data


def _get_data_from_enhanced_image(dset, convert_p):
    from satpy.enhancements.enhancer import get_enhanced_image

    img = get_enhanced_image(dset)
    if convert_p and img.mode == "P":
        img = _apply_palette_to_image(img)
    if img.mode != "P":
        data = img.data.clip(0.0, 1.0)
    else:
        data = img.data
    return data


def _apply_palette_to_image(img):
    if len(img.palette[0]) == 3:
        img = img.convert("RGB")
    elif len(img.palette[0]) == 4:
        img = img.convert("RGBA")
    return img


def add_bands(data: xr.DataArray, bands: Sequence | xr.DataArray) -> xr.DataArray:
    """Add image bands so that input data is compatible with the provided ``bands``.

    This function will add additional copies of input data arrays to allow the
    number of elements in the "bands" dimension to be more compatible with
    the desired number of bands from the ``bands`` argument. This does not
    always mean producing the exact number of bands. For example, if the
    input data is an "L" image with an alpha channel (``["L", "A"]``) and the
    desired bands are ``["R", "G", "B"]`` then the result will be
    ``["R", "G", "B", "A"]`` where the R, G, and B data are copies of the original
    L band data and the A channel is copied from the input.

    If an alpha channel needs to be added to match the desired set of bands then
    a fully opaque alpha channel is created.


    Args:
        data: Input xarray DataArray with a "bands" dimension and coordinate.
            The DataArray should also have a "mode" attribute in ``.attrs``
            with the image mode of the data. So if the "bands" coordinate
            variable is ``["R", "G", "B"]`` then ``.attrs["mode"]`` will
            be ``"RGB"``.
        bands: The bands that the result should be compatible with. If taken
            from another data DataArray via ``.coords["bands"]`` then this
            will be a ``DataArray``. Otherwise an sequence supporting
            ``__contains__`` should be provided. Elements of this argument
            should be individual string characters, one for each band.

    """
    desired_bands = bands.values.tolist() if isinstance(bands, xr.DataArray) else bands
    # Add R, G and B bands, remove L band
    data = _check_mode_p(data, desired_bands)
    data = _check_mode_l(data, desired_bands)
    # Add alpha band
    data = _check_alpha_band(data, desired_bands)
    return data


def _check_mode_p(data: xr.DataArray, bands: Sequence) -> xr.DataArray:
    if "P" in data["bands"].data or "P" in bands:
        raise NotImplementedError("Cannot mix datasets of mode P with other datasets at the moment.")
    return data


def _check_mode_l(data: xr.DataArray, bands: Sequence) -> xr.DataArray:
    new_data: tuple[xr.DataArray, ...]

    if "L" in data["bands"].data and "R" in bands:
        lum = data.sel(bands="L")
        # Keep 'A' if it was present
        if "A" in data["bands"]:
            alpha = data.sel(bands="A")
            new_data = (lum, lum, lum, alpha)
            new_bands = ["R", "G", "B", "A"]
            mode = "RGBA"
        else:
            new_data = (lum, lum, lum)
            new_bands = ["R", "G", "B"]
            mode = "RGB"
        data = xr.concat(new_data, dim="bands", coords={"bands": new_bands})
        data["bands"] = new_bands
        data.attrs["mode"] = mode
    return data


def _check_alpha_band(data: xr.DataArray, bands: Sequence) -> xr.DataArray:
    if "A" not in data["bands"].data and "A" in bands:
        new_data = [data.sel(bands=band) for band in data["bands"].data]
        # Create alpha band based on a copy of the first "real" band
        alpha = new_data[0].copy()
        # WARN: This will be all 1s as floats, but this is not accurate for
        # integer data. Integer data needing an alpha channel is not currently
        # supported.
        alpha.data = da.ones(
            (data.sizes["y"], data.sizes["x"]),
            dtype=new_data[0].dtype,
            chunks=new_data[0].chunks,
        )
        # Rename band to indicate it's alpha
        alpha["bands"] = "A"
        new_data.append(alpha)
        new_data_arr = xr.concat(new_data, dim="bands")
        new_data_arr.attrs["mode"] = data.attrs["mode"] + "A"
        data = new_data_arr
    return data


class SingleBandCompositor(CompositeBase):
    """Basic single-band composite builder.

    This preserves all the attributes of the dataset it is derived from.
    """

    @staticmethod
    def _update_missing_metadata(existing_attrs, new_attrs):
        for key, val in new_attrs.items():
            if key not in existing_attrs and val is not None:
                existing_attrs[key] = val

    def __call__(self, projectables, nonprojectables=None, **attrs):
        """Build the composite."""
        if len(projectables) != 1:
            raise ValueError("Can't have more than one band in a single-band composite")

        data = projectables[0]
        new_attrs = data.attrs.copy()
        self._update_missing_metadata(new_attrs, attrs)
        resolution = new_attrs.get("resolution", None)
        new_attrs.update(self.attrs)
        if resolution is not None:
            new_attrs["resolution"] = resolution

        return xr.DataArray(data=data.data, attrs=new_attrs,
                            dims=data.dims, coords=data.coords)


class GenericCompositor(CompositeBase):
    """Basic colored composite builder."""

    modes = {1: "L", 2: "LA", 3: "RGB", 4: "RGBA"}

    def __init__(self, name, common_channel_mask=True, **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            common_channel_mask (bool): If True, mask all the channels with
                a mask that combines all the invalid areas of the given data.

        """
        self.common_channel_mask = common_channel_mask
        super(GenericCompositor, self).__init__(name, **kwargs)

    @classmethod
    def infer_mode(cls, data_arr):
        """Guess at the mode for a particular DataArray."""
        if "mode" in data_arr.attrs:
            return data_arr.attrs["mode"]
        if "bands" not in data_arr.dims:
            return cls.modes[1]
        if "bands" in data_arr.coords and isinstance(data_arr.coords["bands"][0].item(), str):
            return "".join(data_arr.coords["bands"].values)
        return cls.modes[data_arr.sizes["bands"]]

    def _concat_datasets(self, projectables, mode):
        try:
            data = xr.concat(projectables, "bands", coords="minimal")
            data["bands"] = list(mode)
        except ValueError as e:
            LOG.debug("Original exception for incompatible areas: {}".format(str(e)))
            raise IncompatibleAreas("Areas do not match.")

        return data

    def _get_sensors(self, projectables):
        sensor = set()
        for projectable in projectables:
            current_sensor = projectable.attrs.get("sensor", None)
            if current_sensor:
                if isinstance(current_sensor, (str, bytes)):
                    sensor.add(current_sensor)
                else:
                    sensor |= current_sensor
        if len(sensor) == 0:
            sensor = None
        elif len(sensor) == 1:
            sensor = list(sensor)[0]
        return sensor

    def __call__(
            self,
            datasets: Sequence[xr.DataArray],
            optional_datasets: Optional[Sequence[xr.DataArray]] = None,
            **attrs
    ) -> xr.DataArray:
        """Build the composite."""
        if "deprecation_warning" in self.attrs:
            warnings.warn(
                self.attrs["deprecation_warning"],
                UserWarning,
                stacklevel=2
            )
            self.attrs.pop("deprecation_warning", None)

        mode = self._get_mode(attrs, len(datasets))

        if len(datasets) > 1:
            datasets, data = self._check_datasets_and_data(datasets, mode)
        else:
            data = datasets[0]

        new_attrs = self._get_updated_attrs(datasets, attrs, mode)

        return xr.DataArray(data=data.data, attrs=new_attrs,
                            dims=data.dims, coords=data.coords)

    def _get_mode(self, attrs, num):
        mode = attrs.get("mode")
        if mode is None:
            # num may not be in `self.modes` so only check if we need to
            mode = self.modes[num]
        return mode

    def _check_datasets_and_data(self, datasets, mode):
        datasets = self.match_data_arrays(datasets)
        data = self._concat_datasets(datasets, mode)
        # Skip masking if user wants it or a specific alpha channel is given.
        if self.common_channel_mask and mode[-1] != "A":
            data = data.where(data.notnull().all(dim="bands"))
        # if inputs have a time coordinate that may differ slightly between
        # themselves then find the mid time and use that as the single
        # time coordinate value
        time = check_times(datasets)
        if time is not None and "time" in data.dims:
            data["time"] = [time]

        return datasets, data

    def _get_updated_attrs(self, datasets, attrs, mode):
        new_attrs = combine_metadata(*datasets)
        # remove metadata that shouldn't make sense in a composite
        new_attrs["wavelength"] = None
        new_attrs.pop("units", None)
        new_attrs.pop("calibration", None)
        new_attrs.pop("modifiers", None)

        new_attrs.update({key: val
                          for (key, val) in attrs.items()
                          if val is not None})
        resolution = new_attrs.get("resolution", None)
        new_attrs.update(self.attrs)
        if resolution is not None:
            new_attrs["resolution"] = resolution
        new_attrs["sensor"] = self._get_sensors(datasets)
        new_attrs["mode"] = mode

        return new_attrs


def check_times(projectables):
    """Check that *projectables* have compatible times."""
    times = []
    for proj in projectables:
        status = _collect_time_from_proj(times, proj)
        if not status:
            break
    else:
        return _get_average_time(times)


def _collect_time_from_proj(times, proj):
    status = False
    try:
        if proj["time"].size and proj["time"][0] != 0:
            times.append(proj["time"][0].values)
            status = True
    except KeyError:
        # the datasets don't have times
        pass
    except IndexError:
        # time is a scalar
        if proj["time"].values != 0:
            times.append(proj["time"].values)
            status = True
    return status


def _get_average_time(times):
    # Is there a more gracious way to handle this ?
    if np.max(times) - np.min(times) > TIME_COMPATIBILITY_TOLERANCE:
        raise IncompatibleTimes("Times do not match.")
    return (np.max(times) - np.min(times)) / 2 + np.min(times)


class RGBCompositor(GenericCompositor):
    """Make a composite from three color bands (deprecated)."""

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        warnings.warn(
            "RGBCompositor is deprecated, use GenericCompositor instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if len(projectables) != 3:
            raise ValueError("Expected 3 datasets, got %d" % (len(projectables),))
        return super(RGBCompositor, self).__call__(projectables, **info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the hrpt reader."""

from datetime import datetime, timedelta
from unittest import mock

import numpy as np
import pytest
import xarray as xr

from satpy.readers import hrpt
from satpy.readers.hrpt import HRPTFile, scanline_dtype, time_seconds
from satpy.tests.utils import make_dataid

NUMBER_OF_SCANS = 10
SWATH_WIDTH = 2048

COUNTS = (np.arange(5 * NUMBER_OF_SCANS * SWATH_WIDTH) % 500 + 450).reshape((NUMBER_OF_SCANS, SWATH_WIDTH, 5))
LONS = np.ones((NUMBER_OF_SCANS, SWATH_WIDTH))
LATS = np.ones((NUMBER_OF_SCANS, SWATH_WIDTH)) * 2
START_TIME = datetime(2009, 6, 9, 9, 45)


@pytest.fixture
def hrpt_file(tmp_path):
        """Set up the test case."""
        test_data = np.ones(NUMBER_OF_SCANS, dtype=scanline_dtype)
        time_code = []
        offset = timedelta(seconds=1/6)
        for line in range(NUMBER_OF_SCANS):
            time_code.append(to_timecode(START_TIME + offset * line)[0])
        test_data["timecode"] = time_code
        test_data["image_data"] = COUNTS
        test_data["telemetry"]["PRT"] = 250
        test_data["telemetry"]["PRT"][::5] = 0
        # Channel 3a
        test_data["id"]["id"][:5] = 891
        test_data[:5]["space_data"] = [38, 39, 38, 991, 995]
        test_data[:5]["back_scan"] = [40, 445, 420]
        # Channel 3b
        test_data["id"]["id"][5:] = 890
        test_data[5:]["space_data"] = [38, 39, 980, 991, 995]
        test_data[5:]["back_scan"] = [440, 445, 420]
        filename = tmp_path / "20250609094500_noaa19.hmf"
        test_data.tofile(filename)
        return filename


@pytest.fixture
def hrpt_filename_info():
    """Create expected filename info without actually parsing the filename."""
    return {"start_time": START_TIME}


@pytest.fixture
def hrpt_fh(hrpt_file, hrpt_filename_info):
    """Open the file handler."""
    return HRPTFile(hrpt_file, hrpt_filename_info, {})


class TestHRPTReading:
    """Test case for reading hrpt data."""

    def test_reading(self, hrpt_file, hrpt_filename_info):
        """Test that data is read."""
        fh = HRPTFile(hrpt_file, hrpt_filename_info, {})
        assert fh._data is not None


class TestHRPTGetUncalibratedData:
    """Test case for reading uncalibrated hrpt data."""

    def test_get_dataset_returns_a_dataarray(self, hrpt_fh):
        """Test that get_dataset returns a data array."""
        result = hrpt_fh.get_dataset(make_dataid(name="1", calibration="counts"), {})
        assert isinstance(result, xr.DataArray)

    def test_platform_name(self, hrpt_fh):
        """Test that the platform name is correct."""
        result = hrpt_fh.get_dataset(make_dataid(name="1", calibration="counts"), {})
        assert result.attrs["platform_name"] == "NOAA 19"

    def test_no_calibration_values_are_raw(self, hrpt_fh):
        """Test that the values of uncalibrated data is the raw data."""
        result = hrpt_fh.get_dataset(make_dataid(name="1", calibration="counts"), {})
        assert (result.values == COUNTS[:, :, 0]).all()


class TestHRPTGetCalibratedReflectances:
    """Test case for reading calibrated reflectances from hrpt data."""

    def test_calibrated_reflectances_values(self, hrpt_fh):
        """Test the calibrated reflectance values."""
        result = hrpt_fh.get_dataset(make_dataid(name="1", calibration="reflectance"), {})
        np.testing.assert_allclose(result.values.mean(), 57.772733)


class TestHRPTGetCalibratedBT:
    """Test case for reading calibrated brightness temperature from hrpt data."""

    def test_calibrated_bt_values(self, hrpt_fh):
        """Test the calibrated reflectance values."""
        result = hrpt_fh.get_dataset(make_dataid(name="4", calibration="brightness_temperature"), {})
        np.testing.assert_allclose(result.values.mean(), 249.52884)


class TestHRPTChannel3:
    """Test case for reading calibrated brightness temperature from hrpt data."""

    def test_channel_3b_masking(self, hrpt_fh):
        """Test that channel 3b is split correctly."""
        result = hrpt_fh.get_dataset(make_dataid(name="3b", calibration="brightness_temperature"), {})
        assert np.isnan(result.values[:5]).all()
        assert np.isfinite(result.values[5:]).all()

    def test_channel_3a_masking(self, hrpt_fh):
        """Test that channel 3a is split correctly."""
        result = hrpt_fh.get_dataset(make_dataid(name="3a", calibration="reflectance"), {})
        assert np.isnan(result.values[5:]).all()
        assert np.isfinite(result.values[:5]).all()

    def test_uncalibrated_channel_3a_masking(self, hrpt_fh):
        """Test that channel 3a is split correctly."""
        result = hrpt_fh.get_dataset(make_dataid(name="3a", calibration="counts"), {})
        assert np.isnan(result.values[5:]).all()
        assert np.isfinite(result.values[:5]).all()


@pytest.fixture
def mock_nav(monkeypatch):
        """Prepare the mocks."""
        monkeypatch.setattr(hrpt, "compute_pixels", mock.Mock())
        Orbital = mock.Mock()
        Orbital.return_value.get_position.return_value = mock.MagicMock(), mock.MagicMock()
        monkeypatch.setattr(hrpt, "Orbital", Orbital)
        get_lonlatalt = mock.Mock()
        get_lonlatalt.return_value = (mock.MagicMock(), mock.MagicMock(), mock.MagicMock())
        monkeypatch.setattr(hrpt, "get_lonlatalt", get_lonlatalt)
        SatelliteInterpolator = mock.Mock()
        SatelliteInterpolator.return_value.interpolate.return_value = LONS, LATS
        monkeypatch.setattr(hrpt, "SatelliteInterpolator", SatelliteInterpolator)


class TestHRPTNavigation:
    """Test case for computing HRPT navigation."""

    def test_longitudes_are_returned(self, hrpt_fh, mock_nav):
        """Check that latitudes are returned properly."""
        dataset_id = make_dataid(name="longitude")
        result = hrpt_fh.get_dataset(dataset_id, {})
        assert (result == LONS).all()

    def test_latitudes_are_returned(self, hrpt_fh, mock_nav):
        """Check that latitudes are returned properly."""
        dataset_id = make_dataid(name="latitude")
        result = hrpt_fh.get_dataset(dataset_id, {})
        assert (result == LATS).all()


def to_timecode(dt_time):
    """Convert a datetime to timecode for hrpt scans."""
    year = dt_time.year
    delta = dt_time - datetime(year, 1, 1)
    days = delta.days + 1
    timecode = np.array([0, 0, 0, 0])
    timecode[0] = days << 1
    msecs = int(delta.seconds * 1000 + delta.microseconds / 1000)
    timecode[1] = (msecs >> 20) & 127
    timecode[2] = (msecs >> 10) & 1023
    timecode[3] = msecs & 1023
    return timecode, year


def test_time_seconds():
    """Test conversion of timecode to datetime64."""
    current = datetime.now()
    current = current.replace(microsecond=round(current.microsecond, -3))
    timecode, year = to_timecode(current)
    assert time_seconds(np.array([timecode], "u2"), year) == np.datetime64(current)



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Enhancements based on convolution."""

from __future__ import annotations

import logging

import dask.array as da
import numpy as np

from .wrappers import exclude_alpha, on_dask_array, on_separate_bands

LOG = logging.getLogger(__name__)


def three_d_effect(img, **kwargs):
    """Create 3D effect using convolution."""
    w = kwargs.get("weight", 1)
    LOG.debug("Applying 3D effect with weight %.2f", w)
    kernel = np.array([[-w, 0, w],
                       [-w, 1, w],
                       [-w, 0, w]])
    mode = kwargs.get("convolve_mode", "same")
    return _three_d_effect(img.data, kernel=kernel, mode=mode)


@exclude_alpha
@on_separate_bands
@on_dask_array
def _three_d_effect(band_data, kernel=None, mode=None, index=None):
    del index

    new_data = da.map_blocks(
        _three_d_effect_numpy,
        band_data.rechunk(band_data.shape),
        kernel,
        mode,
        dtype=band_data.dtype,
        meta=np.ndarray((), dtype=band_data.dtype),
    )
    return new_data.rechunk(band_data.chunks)


def _three_d_effect_numpy(band_data, kernel, mode):
    """Kernel for running delayed 3D effect creation."""
    from scipy.signal import convolve2d
    band_data = band_data.reshape(band_data.shape[1:])
    new_data = convolve2d(band_data, kernel, mode=mode)
    return new_data.reshape((1, band_data.shape[0], band_data.shape[1]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Base HDF-EOS reader."""

from __future__ import annotations

import datetime as dt
import logging
import re
from ast import literal_eval
from contextlib import suppress
from functools import cache

import dask.array as da
import numpy as np
import xarray as xr
from pyhdf.error import HDF4Error
from pyhdf.SD import SD

from satpy import DataID
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import normalize_low_res_chunks

logger = logging.getLogger(__name__)


def interpolate(clons, clats, csatz, src_resolution, dst_resolution):
    """Interpolate two parallel datasets jointly."""
    if csatz is None:
        return _interpolate_no_angles(clons, clats, src_resolution, dst_resolution)
    return _interpolate_with_angles(clons, clats, csatz, src_resolution, dst_resolution)


def _interpolate_with_angles(clons, clats, csatz, src_resolution, dst_resolution):
    from geotiepoints.modisinterpolator import modis_1km_to_250m, modis_1km_to_500m, modis_5km_to_1km

    # (src_res, dst_res, is satz not None) -> interp function
    interpolation_functions = {
        (5000, 1000): modis_5km_to_1km,
        (1000, 500): modis_1km_to_500m,
        (1000, 250): modis_1km_to_250m
    }
    return _find_and_run_interpolation(interpolation_functions, src_resolution, dst_resolution,
                                       (clons, clats, csatz))


def _interpolate_no_angles(clons, clats, src_resolution, dst_resolution):
    interpolation_functions = {}

    try:
        from geotiepoints.simple_modis_interpolator import modis_1km_to_250m as simple_1km_to_250m
        from geotiepoints.simple_modis_interpolator import modis_1km_to_500m as simple_1km_to_500m
    except ImportError:
        raise NotImplementedError(
            f"Interpolation from {src_resolution}m to {dst_resolution}m "
            "without satellite zenith angle information is not "
            "implemented. Try updating your version of "
            "python-geotiepoints.")
    else:
        interpolation_functions[(1000, 500)] = simple_1km_to_500m
        interpolation_functions[(1000, 250)] = simple_1km_to_250m

    return _find_and_run_interpolation(interpolation_functions, src_resolution, dst_resolution,
                                       (clons, clats))


def _find_and_run_interpolation(interpolation_functions, src_resolution, dst_resolution, args):
    try:
        interpolation_function = interpolation_functions[(src_resolution, dst_resolution)]
    except KeyError:
        error_message = "Interpolation from {}m to {}m not implemented".format(
            src_resolution, dst_resolution)
        raise NotImplementedError(error_message)

    logger.debug("Interpolating from {} to {}".format(src_resolution, dst_resolution))
    return interpolation_function(*args)

def _modis_date(date):
    """Transform a date and time string into a datetime object."""
    if len(date) == 19:
        return dt.datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
    else:
        return dt.datetime.strptime(date, "%Y-%m-%d %H:%M:%S.%f")

class HDFEOSBaseFileReader(BaseFileHandler):
    """Base file handler for HDF EOS data for both L1b and L2 products."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize the base reader."""
        BaseFileHandler.__init__(self, filename, filename_info, filetype_info)
        try:
            self.sd = SD(self.filename)
        except HDF4Error as err:
            error_message = "Could not load data from file {}: {}".format(self.filename, err)
            raise ValueError(error_message)

        self.metadata = self._load_all_metadata_attributes()

    def _load_all_metadata_attributes(self):
        metadata = {}
        attrs = self.sd.attributes()
        for md_key in ("CoreMetadata.0", "StructMetadata.0", "ArchiveMetadata.0"):
            try:
                str_val = attrs[md_key]
            except KeyError:
                continue
            else:
                metadata.update(self.read_mda(str_val))
        return metadata

    @classmethod
    def read_mda(cls, attribute):
        """Read the EOS metadata."""
        line_iterator = iter(attribute.split("\n"))
        return cls._read_mda(line_iterator)

    @classmethod
    def _read_mda(cls, lines, element=None):
        current_dict = {}

        for line in lines:
            if not line:
                continue
            if line == "END":
                return current_dict

            key, val = cls._split_line(line, lines)

            if key in ["GROUP", "OBJECT"]:
                current_dict[val] = cls._read_mda(lines, val)
            elif key in ["END_GROUP", "END_OBJECT"]:
                if val != element:
                    raise SyntaxError("Non-matching end-tag")
                return current_dict
            elif key in ["CLASS", "NUM_VAL"]:
                pass
            else:
                current_dict[key] = val
        logger.warning("Malformed EOS metadata, missing an END.")
        return current_dict

    @classmethod
    def _split_line(cls, line, lines):
        key, val = line.split("=", maxsplit=1)
        key = key.strip()
        val = val.strip()
        try:
            with suppress(ValueError):
                val = literal_eval(val)
        except SyntaxError:
            key, val = cls._split_line(line + next(lines), lines)
        return key, val

    @property
    def metadata_platform_name(self):
        """Platform name from the internal file metadata."""
        try:
            # Example: 'Terra' or 'Aqua'
            return self.metadata["INVENTORYMETADATA"]["ASSOCIATEDPLATFORMINSTRUMENTSENSOR"][
                "ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER"]["ASSOCIATEDPLATFORMSHORTNAME"]["VALUE"]
        except KeyError:
            return self._platform_name_from_filename()

    def _platform_name_from_filename(self):
        platform_indicator = self.filename_info["platform_indicator"]
        if platform_indicator in ("t", "O"):
            # t1.* or MOD*
            return "Terra"
        # a1.* or MYD*
        return "Aqua"

    @property
    def start_time(self):
        """Get the start time of the dataset."""
        try:
            date = (self.metadata["INVENTORYMETADATA"]["RANGEDATETIME"]["RANGEBEGINNINGDATE"]["VALUE"] + " " +
                    self.metadata["INVENTORYMETADATA"]["RANGEDATETIME"]["RANGEBEGINNINGTIME"]["VALUE"])
            return _modis_date(date)
        except KeyError:
            return self._start_time_from_filename()

    def _start_time_from_filename(self):
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get the end time of the dataset."""
        try:
            date = (self.metadata["INVENTORYMETADATA"]["RANGEDATETIME"]["RANGEENDINGDATE"]["VALUE"] + " " +
                    self.metadata["INVENTORYMETADATA"]["RANGEDATETIME"]["RANGEENDINGTIME"]["VALUE"])
            return _modis_date(date)
        except KeyError:
            return self.start_time

    def _read_dataset_in_file(self, dataset_name):
        if dataset_name not in self.sd.datasets():
            error_message = "Dataset name {} not included in available datasets {}".format(
                dataset_name, self.sd.datasets()
            )
            raise KeyError(error_message)

        dataset = self.sd.select(dataset_name)
        return dataset

    def load_dataset(self, dataset_name, is_category=False):
        """Load the dataset from HDF EOS file."""
        from satpy.readers.core.hdf4 import from_sds

        dataset = self._read_dataset_in_file(dataset_name)
        chunks = self._chunks_for_variable(dataset)
        dask_arr = from_sds(dataset, self.filename, chunks=chunks)
        dims = ("y", "x") if dask_arr.ndim == 2 else None
        data = xr.DataArray(dask_arr, dims=dims,
                            attrs=dataset.attributes())
        data = _scale_and_mask_data_array(data, is_category=is_category)

        return data

    def _chunks_for_variable(self, hdf_dataset):
        scan_length_250m = 40
        var_shape = hdf_dataset.info()[2]
        res_multiplier = self._get_res_multiplier(var_shape)
        num_nonyx_dims = len(var_shape) - 2
        return normalize_low_res_chunks(
            (1,) * num_nonyx_dims + ("auto", -1),
            var_shape,
            (1,) * num_nonyx_dims + (scan_length_250m, var_shape[-1]),
            (1,) * num_nonyx_dims + (res_multiplier, res_multiplier),
            np.float32,
        )

    @staticmethod
    def _get_res_multiplier(var_shape):
        num_columns_to_multiplier = {
            271: 20,  # 5km
            1354: 4,  # 1km
            2708: 2,  # 500m
            5416: 1,  # 250m
        }
        for max_columns, res_multiplier in num_columns_to_multiplier.items():
            if var_shape[-1] <= max_columns:
                return res_multiplier
        return 1

    def _add_satpy_metadata(self, data_id: DataID, data_arr: xr.DataArray):
        """Add metadata that is specific to Satpy."""
        new_attrs = {
            "platform_name": self.metadata_platform_name,
            "sensor": "modis",
        }

        res = data_id["resolution"]
        rps = self._resolution_to_rows_per_scan(res)
        new_attrs["rows_per_scan"] = rps

        data_arr.attrs.update(new_attrs)

    def _resolution_to_rows_per_scan(self, resolution: int) -> int:
        known_rps = {
            5000: 2,
            1000: 10,
            500: 20,
            250: 40,
        }
        return known_rps.get(resolution, 10)


class HDFEOSGeoReader(HDFEOSBaseFileReader):
    """Handler for the geographical datasets."""

    # list of geographical datasets handled by the georeader
    # mapping to the default variable name if not specified in YAML
    DATASET_NAMES = {
        "longitude": "Longitude",
        "latitude": "Latitude",
        "satellite_azimuth_angle": ("SensorAzimuth", "Sensor_Azimuth"),
        "satellite_zenith_angle": ("SensorZenith", "Sensor_Zenith"),
        "solar_azimuth_angle": ("SolarAzimuth", "SolarAzimuth"),
        "solar_zenith_angle": ("SolarZenith", "Solar_Zenith"),
        "water_present": "WaterPresent",
        "landsea_mask": "Land/SeaMask",
        "height": "Height",
        "range": "Range",
    }

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize the geographical reader."""
        HDFEOSBaseFileReader.__init__(self, filename, filename_info, filetype_info, **kwargs)
        self._load_interpolated_lonlat_pair = cache(self._load_interpolated_lonlat_pair_uncached)
        self._load_interpolated_angle_pair = cache(self._load_interpolated_angle_pair_uncached)

    @staticmethod
    def is_geo_loadable_dataset(dataset_name: str) -> bool:
        """Determine if this dataset should be loaded as a Geo dataset."""
        return dataset_name in HDFEOSGeoReader.DATASET_NAMES

    @staticmethod
    def read_geo_resolution(metadata):
        """Parse metadata to find the geolocation resolution."""
        # level 1 files
        try:
            return HDFEOSGeoReader._geo_resolution_for_l1b(metadata)
        except KeyError:
            try:
                return HDFEOSGeoReader._geo_resolution_for_l2_l1b(metadata)
            except (AttributeError, KeyError):
                raise RuntimeError("Could not determine resolution from file metadata")

    @staticmethod
    def _geo_resolution_for_l1b(metadata):
        ds = metadata["INVENTORYMETADATA"]["COLLECTIONDESCRIPTIONCLASS"]["SHORTNAME"]["VALUE"]
        if ds.endswith("D03") or ds.endswith("HKM") or ds.endswith("QKM"):
            return 1000
        # 1km files have 5km geolocation usually
        return 5000

    @staticmethod
    def _geo_resolution_for_l2_l1b(metadata):
        # data files probably have this level 2 files
        # this does not work for L1B 1KM data files because they are listed
        # as 1KM data but the geo data inside is at 5km
        latitude_dim = metadata["SwathStructure"]["SWATH_1"]["DimensionMap"]["DimensionMap_2"]["GeoDimension"]
        resolution_regex = re.compile(r"(?P<resolution>\d+)(km|KM)")
        resolution_match = resolution_regex.search(latitude_dim)
        return int(resolution_match.group("resolution")) * 1000

    @property
    def geo_resolution(self):
        """Resolution of the geographical data retrieved in the metadata."""
        return self.read_geo_resolution(self.metadata)

    def _load_ds_by_name(self, ds_name):
        """Attempt loading using multiple common names."""
        var_names = self.DATASET_NAMES[ds_name]
        if isinstance(var_names, (list, tuple)):
            try:
                return self.load_dataset(var_names[0])
            except KeyError:
                return self.load_dataset(var_names[1])
        return self.load_dataset(var_names)

    def get_dataset(self, dataset_id: DataID, dataset_info: dict) -> xr.DataArray:
        """Get the geolocation dataset."""
        # Name of the dataset as it appears in the HDF EOS file
        in_file_dataset_name = dataset_info.get("file_key")
        # Name of the dataset in the YAML file
        dataset_name = dataset_id["name"]
        # Resolution asked
        resolution = dataset_id["resolution"]
        if in_file_dataset_name is not None:
            # if the YAML was configured with a specific name use that
            data = self.load_dataset(in_file_dataset_name)
        else:
            # otherwise use the default name for this variable
            data = self._load_ds_by_name(dataset_name)
        if resolution != self.geo_resolution:
            if in_file_dataset_name is not None:
                # they specified a custom variable name but
                # we don't know how to interpolate this yet
                raise NotImplementedError(
                    f"Interpolation for variable '{dataset_name}' is not "
                    "configured.")

            logger.debug(f"Loading and interpolating {dataset_name}")
            data = self.get_interpolated_dataset(dataset_name, resolution)

        for key in ("standard_name", "units"):
            if key in dataset_info:
                data.attrs[key] = dataset_info[key]
        self._add_satpy_metadata(dataset_id, data)

        return data

    def get_interpolated_dataset(self, dataset_name: str, resolution: int) -> xr.DataArray:
        """Load and interpolate datasets."""
        interp_pairs = {
            ("longitude", "latitude"): self._load_interpolated_lonlat_pair,
            ("satellite_azimuth_angle", "satellite_zenith_angle"): self._load_interpolated_angle_pair,
            ("solar_azimuth_angle", "solar_zenith_angle"): self._load_interpolated_angle_pair,
        }
        for ds_name_pair, interp_func in interp_pairs.items():
            try:
                pair_index = ds_name_pair.index(dataset_name)
            except ValueError:
                continue
            return interp_func(*ds_name_pair, resolution)[pair_index]
        raise ValueError(f"Dataset {dataset_name} can not be interpolated")

    def _load_interpolated_lonlat_pair_uncached(
            self,
            name1: str,
            name2: str,
            resolution: int
    ) -> tuple[xr.DataArray, xr.DataArray]:
        result1 = self._load_ds_by_name(name1)
        result2 = self._load_ds_by_name(name2)
        return self._interpolate_using_sza(result1, result2, resolution)

    def _load_interpolated_angle_pair_uncached(
            self,
            name1: str,
            name2: str,
            resolution: int
    ) -> tuple[xr.DataArray, xr.DataArray]:
        result1 = self._load_ds_by_name(name1)
        result2 = self._load_ds_by_name(name2) - 90
        interp_result1, interp_result2 = self._interpolate_using_sza(result1, result2, resolution)
        return interp_result1, interp_result2 + 90

    def _interpolate_using_sza(
            self,
            data1: xr.DataArray,
            data2: xr.DataArray,
            resolution: int
    ) -> tuple[xr.DataArray, xr.DataArray]:
        try:
            sensor_zenith = self._load_ds_by_name("satellite_zenith_angle")
        except KeyError:
            # no sensor zenith angle, do "simple" interpolation
            sensor_zenith = None
        return interpolate(
            data1, data2, sensor_zenith,
            self.geo_resolution, resolution
        )


def _scale_and_mask_data_array(data_arr: xr.DataArray, is_category: bool = False) -> xr.DataArray:
    """Unscale byte data and mask invalid/fill values.

    MODIS requires unscaling the in-file bytes in an unexpected way::

        data = (byte_value - add_offset) * scale_factor

    See the below L1B User's Guide Appendix C for more information:

    https://mcst.gsfc.nasa.gov/sites/default/files/file_attachments/M1054E_PUG_2017_0901_V6.2.2_Terra_V6.2.1_Aqua.pdf

    """
    scale_factor = data_arr.attrs.pop("scale_factor", None)
    add_offset = data_arr.attrs.pop("add_offset", None)
    # preserve _FillValue if category
    fill_value = data_arr.attrs.get("_FillValue", None) if is_category else data_arr.attrs.pop("_FillValue", None)
    noncat_dtype = data_arr.dtype.type if np.issubdtype(data_arr.dtype, np.floating) else np.float32
    dtype = data_arr.dtype if is_category else noncat_dtype
    new_data = da.map_blocks(
        _mapblocks_scale_and_mask,
        data_arr.data,
        dtype=dtype,
        meta=np.array((), dtype=dtype),
        token="scale_and_mask",  # nosec: B106
        scale_factor=scale_factor,
        add_offset=add_offset,
        fill_value=fill_value,
        is_category=is_category)
    data_arr = data_arr.copy()
    data_arr.data = new_data
    return data_arr


def _mapblocks_scale_and_mask(arr, scale_factor, add_offset, fill_value, is_category):
    good_mask, new_fill = _get_good_data_mask(arr, fill_value, is_category=is_category)
    # don't scale category products, even though scale_factor may equal 1
    # we still need to convert integers to floats
    if scale_factor is not None and not is_category:
        if add_offset is not None and add_offset != 0:
            arr = arr - np.float32(add_offset)
        arr = arr * np.float32(scale_factor)

    if good_mask is not None:
        arr = np.where(good_mask, arr, new_fill)
    return arr


def _get_good_data_mask(arr, fill_value, is_category=False):
    if fill_value is None:
        return None, None
    # preserve integer data types if possible
    if is_category and np.issubdtype(arr.dtype, np.integer):
        # no need to mask, the fill value is already what it needs to be
        return None, None
    fill_type = arr.dtype.type if np.issubdtype(arr.dtype, np.floating) else np.float32
    new_fill = fill_type(np.nan)
    good_mask = arr != fill_value
    return good_mask, new_fill



"""Unit tests for GMS-5 VISSR reader."""

import datetime as dt
import gzip

import fsspec
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

import satpy.tests.reader_tests.gms.test_gms5_vissr_data as real_world
from satpy.tests.reader_tests.utils import get_jit_methods
from satpy.tests.utils import make_dataid, skip_numba_unstable_if_missing

try:
    import satpy.readers.gms.gms5_vissr_format as fmt
    import satpy.readers.gms.gms5_vissr_l1b as vissr
    import satpy.readers.gms.gms5_vissr_navigation as nav
except ImportError as err:
    if skip_numba_unstable_if_missing():
        pytest.skip(f"Numba is not compatible with unstable NumPy: {err!s}", allow_module_level=True)
    raise


@pytest.fixture(params=[False, True], autouse=True)
def _disable_jit(request, monkeypatch):
    """Run tests with jit enabled and disabled.

    Reason: Coverage report is only accurate with jit disabled.
    """
    if request.param:
        jit_methods = get_jit_methods(vissr)
        for name, method in jit_methods.items():
            monkeypatch.setattr(name, method.py_func)


class TestEarthMask:
    """Test getting the earth mask."""

    def test_get_earth_mask(self):
        """Test getting the earth mask."""
        first_earth_pixels = np.array([-1, 1, 0, -1])
        last_earth_pixels = np.array([-1, 3, 2, -1])
        edges = first_earth_pixels, last_earth_pixels
        mask_exp = np.array(
            [[0, 0, 0, 0],
             [0, 1, 1, 1],
             [1, 1, 1, 0],
             [0, 0, 0, 0]]
        )
        mask = vissr.get_earth_mask(mask_exp.shape, edges)
        np.testing.assert_equal(mask, mask_exp)


class TestFileHandler:
    """Test VISSR file handler."""

    @pytest.fixture(autouse=True)
    def _patch_number_of_pixels_per_scanline(self, monkeypatch):
        """Patch data types so that each scanline has two pixels."""
        num_pixels = 2
        IMAGE_DATA_BLOCK_IR = np.dtype(
            [
                ("LCW", fmt.LINE_CONTROL_WORD),
                ("DOC", fmt.U1, (256,)),
                ("image_data", fmt.U1, num_pixels),
            ]
        )
        IMAGE_DATA_BLOCK_VIS = np.dtype(
            [
                ("LCW", fmt.LINE_CONTROL_WORD),
                ("DOC", fmt.U1, (64,)),
                ("image_data", fmt.U1, (num_pixels,)),
            ]
        )
        IMAGE_DATA = {
            fmt.VIS_CHANNEL: {
                "offset": 6 * fmt.BLOCK_SIZE_VIS,
                "dtype": IMAGE_DATA_BLOCK_VIS,
            },
            fmt.IR_CHANNEL: {
                "offset": 18 * fmt.BLOCK_SIZE_IR,
                "dtype": IMAGE_DATA_BLOCK_IR,
            },
        }
        monkeypatch.setattr(
            "satpy.readers.gms.gms5_vissr_format.IMAGE_DATA_BLOCK_IR", IMAGE_DATA_BLOCK_IR
        )
        monkeypatch.setattr(
            "satpy.readers.gms.gms5_vissr_format.IMAGE_DATA_BLOCK_VIS", IMAGE_DATA_BLOCK_VIS
        )
        monkeypatch.setattr("satpy.readers.gms.gms5_vissr_format.IMAGE_DATA", IMAGE_DATA)

    @pytest.fixture(
        params=[
            make_dataid(name="VIS", calibration="reflectance", resolution=1250),
            make_dataid(
                name="IR1", calibration="brightness_temperature", resolution=5000
            ),
            make_dataid(name="IR1", calibration="counts", resolution=5000),
        ]
    )
    def dataset_id(self, request):
        """Get dataset ID."""
        return request.param

    @pytest.fixture(params=[True, False])
    def mask_space(self, request):
        """Mask space pixels."""
        return request.param

    @pytest.fixture(params=[True, False])
    def with_compression(self, request):
        """Enable compression."""
        return request.param

    @pytest.fixture
    def open_function(self, with_compression):
        """Get open function for writing test files."""
        return gzip.open if with_compression else open

    @pytest.fixture
    def vissr_filename(self, tmp_path, with_compression):
        """Construct the VISSR filename."""
        return tmp_path / ("vissr_file" + (".gz" if with_compression else ""))


    @pytest.fixture
    def vissr_file(self, dataset_id, file_contents, open_function, vissr_filename):
        """Get test VISSR file."""
        ch_type = fmt.CHANNEL_TYPES[dataset_id["name"]]
        writer = VissrFileWriter(ch_type, open_function)
        writer.write(vissr_filename, file_contents)
        return vissr_filename

    @pytest.fixture
    def file_contents(self, control_block, image_parameters, image_data):
        """Get VISSR file contents."""
        return {
            "control_block": control_block,
            "image_parameters": image_parameters,
            "image_data": image_data,
        }

    @pytest.fixture
    def control_block(self, dataset_id):
        """Get VISSR control block."""
        block_size = {"IR1": 16, "VIS": 4}
        ctrl_block = np.zeros(1, dtype=fmt.CONTROL_BLOCK)
        ctrl_block["parameter_block_size"] = block_size[dataset_id["name"]]
        ctrl_block["available_block_size_of_image_data"] = 2
        return ctrl_block

    @pytest.fixture
    def image_parameters(self, mode_block, cal_params, nav_params):
        """Get VISSR image parameters."""
        image_params = {"mode": mode_block}
        image_params.update(cal_params)
        image_params.update(nav_params)
        return image_params

    @pytest.fixture
    def nav_params(
        self,
        coordinate_conversion,
        attitude_prediction,
        orbit_prediction,
    ):
        """Get navigation parameters."""
        nav_params = {}
        nav_params.update(attitude_prediction)
        nav_params.update(orbit_prediction)
        nav_params.update(coordinate_conversion)
        return nav_params

    @pytest.fixture
    def cal_params(
        self,
        vis_calibration,
        ir1_calibration,
        ir2_calibration,
        wv_calibration,
    ):
        """Get calibration parameters."""
        return {
            "vis_calibration": vis_calibration,
            "ir1_calibration": ir1_calibration,
            "ir2_calibration": ir2_calibration,
            "wv_calibration": wv_calibration,
        }

    @pytest.fixture
    def mode_block(self):
        """Get VISSR mode block."""
        mode = np.zeros(1, dtype=fmt.MODE_BLOCK)
        mode["satellite_name"] = b"GMS-5       "
        mode["spin_rate"] = 99.21774
        mode["observation_time_mjd"] = 50000.0
        mode["ssp_longitude"] = 140.0
        mode["satellite_height"] = 123456.0
        mode["ir_frame_parameters"]["number_of_lines"] = 2
        mode["ir_frame_parameters"]["number_of_pixels"] = 2
        mode["vis_frame_parameters"]["number_of_lines"] = 2
        mode["vis_frame_parameters"]["number_of_pixels"] = 2
        return mode

    @pytest.fixture
    def coordinate_conversion(self, coord_conv, simple_coord_conv_table):
        """Get all coordinate conversion parameters."""
        return {
            "coordinate_conversion": coord_conv,
            "simple_coordinate_conversion_table": simple_coord_conv_table
        }

    @pytest.fixture
    def coord_conv(self):
        """Get parameters for coordinate conversions.

        Adjust pixel offset so that the first column is at the image center.
        This has the advantage that we can test with very small 2x2 images.
        Otherwise, all pixels would be in space.
        """
        conv = np.zeros(1, dtype=fmt.COORDINATE_CONVERSION_PARAMETERS)

        cline = conv["central_line_number_of_vissr_frame"]
        cline["IR1"] = 1378.5
        cline["VIS"] = 5513.0

        cpix = conv["central_pixel_number_of_vissr_frame"]
        cpix["IR1"] = 0.5  # instead of 1672.5
        cpix["VIS"] = 0.5  # instead of 6688.5

        conv["scheduled_observation_time"] = 50130.979089568464

        nsensors = conv["number_of_sensor_elements"]
        nsensors["IR1"] = 1
        nsensors["VIS"] = 4

        sampling_angle = conv["sampling_angle_along_pixel"]
        sampling_angle["IR1"] = 9.5719995e-05
        sampling_angle["VIS"] = 2.3929999e-05

        stepping_angle = conv["stepping_angle_along_line"]
        stepping_angle["IR1"] = 0.00014000005
        stepping_angle["VIS"] = 3.5000005e-05

        conv["matrix_of_misalignment"] = np.array(
            [[9.9999917e-01, -5.1195198e-04, -1.2135329e-03],
             [5.1036407e-04, 9.9999905e-01, -1.3083406e-03],
             [1.2142011e-03, 1.3077201e-03, 9.9999845e-01]],
            dtype=np.float32
        )

        conv["parameters"]["equatorial_radius"] = 6377397.0
        conv["parameters"]["oblateness_of_earth"] = 0.003342773

        conv["orbital_parameters"]["longitude_of_ssp"] = 141.0
        conv["orbital_parameters"]["latitude_of_ssp"] = 1.0
        return conv

    @pytest.fixture
    def attitude_prediction(self):
        """Get attitude prediction."""
        att_pred = np.zeros(1, dtype=fmt.ATTITUDE_PREDICTION)
        att_pred["data"] = real_world.ATTITUDE_PREDICTION
        return {"attitude_prediction": att_pred}

    @pytest.fixture
    def orbit_prediction(self, orbit_prediction_1, orbit_prediction_2):
        """Get predictions of orbital parameters."""
        return {
            "orbit_prediction_1": orbit_prediction_1,
            "orbit_prediction_2": orbit_prediction_2
        }

    @pytest.fixture
    def orbit_prediction_1(self):
        """Get first block of orbit prediction data."""
        orb_pred = np.zeros(1, dtype=fmt.ORBIT_PREDICTION)
        orb_pred["data"] = real_world.ORBIT_PREDICTION_1
        return orb_pred

    @pytest.fixture
    def orbit_prediction_2(self):
        """Get second block of orbit prediction data."""
        orb_pred = np.zeros(1, dtype=fmt.ORBIT_PREDICTION)
        orb_pred["data"] = real_world.ORBIT_PREDICTION_2
        return orb_pred

    @pytest.fixture
    def vis_calibration(self):
        """Get VIS calibration block."""
        vis_cal = np.zeros(1, dtype=fmt.VIS_CALIBRATION)
        table = vis_cal["vis1_calibration_table"]["brightness_albedo_conversion_table"]
        table[0, 0:4] = np.array([0, 0.25, 0.5, 1])
        return vis_cal

    @pytest.fixture
    def ir1_calibration(self):
        """Get IR1 calibration block."""
        cal = np.zeros(1, dtype=fmt.IR_CALIBRATION)
        table = cal["conversion_table_of_equivalent_black_body_temperature"]
        table[0, 0:4] = np.array([0, 100, 200, 300])
        return cal

    @pytest.fixture
    def ir2_calibration(self):
        """Get IR2 calibration block."""
        cal = np.zeros(1, dtype=fmt.IR_CALIBRATION)
        return cal

    @pytest.fixture
    def wv_calibration(self):
        """Get WV calibration block."""
        cal = np.zeros(1, dtype=fmt.IR_CALIBRATION)
        return cal

    @pytest.fixture
    def simple_coord_conv_table(self):
        """Get simple coordinate conversion table."""
        table = np.zeros(1, dtype=fmt.SIMPLE_COORDINATE_CONVERSION_TABLE)
        table["satellite_height"] = 123457.0
        return table

    @pytest.fixture
    def image_data(self, dataset_id, image_data_ir1, image_data_vis):
        """Get VISSR image data."""
        data = {"IR1": image_data_ir1, "VIS": image_data_vis}
        return data[dataset_id["name"]]

    @pytest.fixture
    def image_data_ir1(self):
        """Get IR1 image data."""
        image_data = np.zeros(2, fmt.IMAGE_DATA_BLOCK_IR)
        image_data["LCW"]["line_number"] = [686, 2089]
        image_data["LCW"]["scan_time"] = [50000, 50000]
        image_data["LCW"]["west_side_earth_edge"] = [0, 0]
        image_data["LCW"]["east_side_earth_edge"] = [1, 1]
        image_data["image_data"] = [[0, 1], [2, 3]]
        return image_data

    @pytest.fixture
    def image_data_vis(self):
        """Get VIS image data."""
        image_data = np.zeros(2, fmt.IMAGE_DATA_BLOCK_VIS)
        image_data["LCW"]["line_number"] = [2744, 8356]
        image_data["LCW"]["scan_time"] = [50000, 50000]
        image_data["LCW"]["west_side_earth_edge"] = [-1, 0]
        image_data["LCW"]["east_side_earth_edge"] = [-1, 1]
        image_data["image_data"] = [[0, 1], [2, 3]]
        return image_data

    @pytest.fixture
    def vissr_file_like(self, vissr_file, with_compression):
        """Get file-like object for VISSR test file."""
        from satpy.readers.core.remote import FSFile

        if with_compression:
            open_file = fsspec.open(vissr_file, compression="gzip")
            return FSFile(open_file)
        return vissr_file

    @pytest.fixture
    def file_handler(self, vissr_file_like, mask_space):
        """Get file handler to be tested."""
        return vissr.GMS5VISSRFileHandler(
            vissr_file_like, {}, {}, mask_space=mask_space
        )

    @pytest.fixture
    def vis_refl_exp(self, mask_space, lons_lats_exp):
        """Get expected VIS reflectance."""
        lons, lats = lons_lats_exp
        if mask_space:
            data = [[np.nan, np.nan], [50, 100]]
        else:
            data = [[0, 25], [50, 100]]
        return xr.DataArray(
            data,
            dims=("y", "x"),
            coords={
                "lon": lons,
                "lat": lats,
                "acq_time": (
                    "y",
                    [dt.datetime(1995, 10, 10), dt.datetime(1995, 10, 10)],
                ),
                "line_number": ("y", [2744, 8356]),
            },
        )

    @pytest.fixture
    def ir1_counts_exp(self, lons_lats_exp):
        """Get expected IR1 counts."""
        lons, lats = lons_lats_exp
        return xr.DataArray(
            [[0, 1], [2, 3]],
            dims=("y", "x"),
            coords={
                "lon": lons,
                "lat": lats,
                "acq_time": (
                    "y",
                    [dt.datetime(1995, 10, 10), dt.datetime(1995, 10, 10)],
                ),
                "line_number": ("y", [686, 2089]),
            },
        )

    @pytest.fixture
    def ir1_bt_exp(self, lons_lats_exp):
        """Get expected IR1 brightness temperature."""
        lons, lats = lons_lats_exp
        return xr.DataArray(
            [[0, 100], [200, 300]],
            dims=("y", "x"),
            coords={
                "lon": lons,
                "lat": lats,
                "acq_time": (
                    "y",
                    [dt.datetime(1995, 10, 10), dt.datetime(1995, 10, 10)],
                ),
                "line_number": ("y", [686, 2089]),
            },
        )

    @pytest.fixture
    def lons_lats_exp(self, dataset_id):
        """Get expected lon/lat coordinates.

        Computed with JMA's Msial library for 2 pixels near the central column
        (6688.5/1672.5 for VIS/IR).

        VIS:

        pix = [6688, 6688, 6689, 6689]
        lin = [2744, 8356, 2744, 8356]

        IR1:

        pix = [1672, 1672, 1673, 1673]
        lin = [686, 2089, 686, 2089]
        """
        expectations = {
            "IR1": {
                "lons": [[139.680120, 139.718902],
                         [140.307367, 140.346062]],
                "lats": [[35.045132, 35.045361],
                         [-34.971012, -34.970738]]
            },
            "VIS": {
                "lons": [[139.665133, 139.674833],
                         [140.292579, 140.302249]],
                "lats": [[35.076113, 35.076170],
                         [-34.940439, -34.940370]]
            }
        }
        exp = expectations[dataset_id["name"]]
        lons = xr.DataArray(exp["lons"], dims=("y", "x"))
        lats = xr.DataArray(exp["lats"], dims=("y", "x"))
        return lons, lats

    @pytest.fixture
    def dataset_exp(self, dataset_id, ir1_counts_exp, ir1_bt_exp, vis_refl_exp):
        """Get expected dataset."""
        ir1_counts_id = make_dataid(name="IR1", calibration="counts", resolution=5000)
        ir1_bt_id = make_dataid(
            name="IR1", calibration="brightness_temperature", resolution=5000
        )
        vis_refl_id = make_dataid(
            name="VIS", calibration="reflectance", resolution=1250
        )
        expectations = {
            ir1_counts_id: ir1_counts_exp,
            ir1_bt_id: ir1_bt_exp,
            vis_refl_id: vis_refl_exp,
        }
        return expectations[dataset_id]

    @pytest.fixture
    def area_def_exp(self, dataset_id):
        """Get expected area definition."""
        if dataset_id["name"] == "IR1":
            resol = 5
            size = 2366
            extent = (-20438.1468, -20438.1468, 20455.4306, 20455.4306)
        else:
            resol = 1
            size = 9464
            extent = (-20444.6235, -20444.6235, 20448.9445, 20448.9445)
        area_id = f"gms-5_vissr_western-pacific_{resol}km"
        desc = f"GMS-5 VISSR Western Pacific area definition with {resol} km resolution"
        return AreaDefinition(
            area_id=area_id,
            description=desc,
            proj_id=area_id,
            projection={
                "a": nav.EARTH_EQUATORIAL_RADIUS,
                "b": nav.EARTH_POLAR_RADIUS,
                "h": "123456",
                "lon_0": "140",
                "no_defs": "None",
                "proj": "geos",
                "type": "crs",
                "units": "m",
                "x_0": "0",
                "y_0": "0",
            },
            area_extent=extent,
            width=size,
            height=size,
        )

    @pytest.fixture
    def attrs_exp(self, area_def_exp):
        """Get expected dataset attributes."""
        return {
            "yaml": "info",
            "platform": "GMS-5",
            "sensor": "VISSR",
            "time_parameters": {
                "nominal_start_time": dt.datetime(1995, 10, 10),
                "nominal_end_time": dt.datetime(1995, 10, 10, 0, 25),
            },
            "orbital_parameters": {
                "satellite_nominal_longitude": 140.0,
                "satellite_nominal_latitude": 0.0,
                "satellite_nominal_altitude": 123456.0,
                "satellite_actual_longitude": 141.0,
                "satellite_actual_latitude": 1.0,
                "satellite_actual_altitude": 123457.0,
            },
            "area_def_uniform_sampling": area_def_exp,
        }

    def test_get_dataset(self, file_handler, dataset_id, dataset_exp, attrs_exp):
        """Test getting the dataset."""
        dataset = file_handler.get_dataset(dataset_id, {"yaml": "info"})
        xr.testing.assert_allclose(dataset.compute(), dataset_exp, atol=1e-6)
        assert dataset.attrs == attrs_exp

    def test_time_attributes(self, file_handler, attrs_exp):
        """Test the file handler's time attributes."""
        start_time_exp = attrs_exp["time_parameters"]["nominal_start_time"]
        end_time_exp = attrs_exp["time_parameters"]["nominal_end_time"]
        assert file_handler.start_time == start_time_exp
        assert file_handler.end_time == end_time_exp


class TestCorruptFile:
    """Test reading corrupt files."""

    @pytest.fixture
    def file_contents(self):
        """Get corrupt file contents (all zero)."""
        control_block = np.zeros(1, dtype=fmt.CONTROL_BLOCK)
        image_data = np.zeros(1, dtype=fmt.IMAGE_DATA_BLOCK_IR)
        return {
            "control_block": control_block,
            "image_parameters": {},
            "image_data": image_data,
        }

    @pytest.fixture
    def corrupt_file(self, file_contents, tmp_path):
        """Write corrupt VISSR file to disk."""
        filename = tmp_path / "my_vissr_file"
        writer = VissrFileWriter(ch_type="VIS", open_function=open)
        writer.write(filename, file_contents)
        return filename

    def test_corrupt_file(self, corrupt_file):
        """Test reading a corrupt file."""
        with pytest.raises(ValueError, match=r".* corrupt .*"):
            vissr.GMS5VISSRFileHandler(corrupt_file, {}, {})


class VissrFileWriter:
    """Write data in VISSR archive format."""

    image_params_order = [
        "mode",
        "coordinate_conversion",
        "attitude_prediction",
        "orbit_prediction_1",
        "orbit_prediction_2",
        "vis_calibration",
        "ir1_calibration",
        "ir2_calibration",
        "wv_calibration",
        "simple_coordinate_conversion_table",
    ]

    def __init__(self, ch_type, open_function):
        """Initialize the writer.

        Args:
            ch_type: Channel type (VIS or IR)
            open_function: Open function to be used (e.g. open or gzip.open)
        """
        self.ch_type = ch_type
        self.open_function = open_function

    def write(self, filename, contents):
        """Write file contents to disk."""
        with self.open_function(filename, mode="wb") as fd:
            self._write_control_block(fd, contents)
            self._write_image_parameters(fd, contents)
            self._write_image_data(fd, contents)

    def _write_control_block(self, fd, contents):
        self._write(fd, contents["control_block"])

    def _write_image_parameters(self, fd, contents):
        for name in self.image_params_order:
            im_param = contents["image_parameters"].get(name)
            if im_param:
                self._write_image_parameter(fd, im_param, name)

    def _write_image_parameter(self, fd, im_param, name):
        offset = fmt.IMAGE_PARAMS[name]["offset"][self.ch_type]
        self._write(fd, im_param, offset)

    def _write_image_data(self, fd, contents):
        offset = fmt.IMAGE_DATA[self.ch_type]["offset"]
        self._write(fd, contents["image_data"], offset)

    def _write(self, fd, data, offset=None):
        """Write data to file.

        If specified, prepend with 'offset' placeholder bytes.
        """
        if offset:
            self._fill(fd, offset)
        fd.write(data.tobytes())

    def _fill(self, fd, target_byte):
        """Write placeholders from current position to target byte."""
        nbytes = target_byte - fd.tell()
        fd.write(b" " * nbytes)
